import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from ..constants import xyz_dict, xyz_list
from ..utils import listify, log_and_raise
from ..monitor import TimeMonitor

from .structure import _plot_eps

def _monitor_slice(sim, mnt_data, sind=None, normal=None, normal_ind=0,
                    field='E', val='re', comp='z', eps_alpha=0):

    # Import in order to distinguish in the axis title
    from ..monitor import TimeMonitor, FreqMonitor

    if normal is None:
        dmin = np.argmin(mnt_data.inds_end - mnt_data.inds_beg)
        normal = ['x', 'y', 'z'][dmin]

    inds = {'x': 0, 'y': 1, 'z': 2}
    comp_ind = inds[comp]
    if field.lower()=='e':
        ftmp = mnt_data.E[:, :, :, :, sind]
        ftit = 'E'
    elif field.lower()=='h':
        ftmp = mnt_data.H[:, :, :, :, sind]
        ftit = 'H'
    elif field.lower()=='s':
        ftmp = mnt_data.S[:, :, :, :, sind]
        ftit = 'S'

    if isinstance(mnt_data.monitor, TimeMonitor):
        tit_string = 't='+"%1.2e"%(mnt_data.tmesh[sind]*1e12)+'fs'
    elif isinstance(mnt_data.monitor, FreqMonitor):
        tit_string = 'f='+"%1.2e"%(mnt_data.freqs[sind]*1e-12)+'THz'

    if val=='int':
        fvals = ftmp[0, :, :, :]**2 + ftmp[1, :, :, :]**2 + ftmp[2, :, :, :]**2
        ax_tit = "||%s||^2"%ftit
    else:
        fvals = ftmp[comp_ind, :, :, :]
        ax_tit = val + "(" + ftit +  comp + ")"

    mbeg = mnt_data.inds_beg
    mend = mnt_data.inds_end

    if normal=='x':
        fvals = fvals[normal_ind, :, :]
        norm_pos = sim.grid.mesh[0][mbeg[0]+normal_ind]
        mesh = (np.array([norm_pos]), sim.grid.mesh[1][mbeg[1]:mend[1]],
                    sim.grid.mesh[2][mbeg[2]:mend[2]])
    elif normal=='y':
        fvals = fvals[:, normal_ind, :]
        norm_pos = sim.grid.mesh[1][mbeg[1]+normal_ind]
        mesh = (sim.grid.mesh[0][mbeg[0]:mend[0]], np.array([norm_pos]),
                    sim.grid.mesh[2][mbeg[2]:mend[2]])
    elif normal=='z':
        fvals = fvals[:, :, normal_ind]
        norm_pos = sim.grid.mesh[2][mbeg[2]+normal_ind]
        mesh = (sim.grid.mesh[0][mbeg[0]:mend[0]],
                    sim.grid.mesh[1][mbeg[1]:mend[1]], np.array([norm_pos]))

    if eps_alpha<=0:
        eps_r = None
    else:
        eps_r = np.squeeze(sim._get_eps(mesh))
        
    grid_dict = {
            'x': (1, 2, 0, 'y', 'z'),
            'y': (0, 2, 1, 'x', 'z'),
            'z': (0, 1, 2, 'x', 'y'),
            }

    (d1, d2, dn, x_lab, y_lab) = grid_dict[normal]
    grid1 = sim.grid.mesh[d1][mbeg[d1]:mend[d1]]
    grid2 = sim.grid.mesh[d2][mbeg[d2]:mend[d2]]
    extent = [grid1[0], grid1[-1], grid2[0], grid2[-1]]

    ax_tit += ', ' + tit_string
    ax_title = 'Monitor ' + mnt_data.name + ', ' + normal + '=' + \
                '%1.2eum\n'%norm_pos + ax_tit

    return (fvals, eps_r, extent, x_lab, y_lab, ax_title)

def _plot_field_2D(fvals, eps_r, extent, x_lab, y_lab, ax_title,
                    val, ax, cbar, clim, eps_alpha):

    cmap = "RdBu"
    cmap_eps = "Greys"
    if val=='re':
        fvals = np.real(fvals)
    elif val=='im':
        fvals = np.imag(fvals)
    else:
        cmap = "magma"
        cmap_eps = "Greys_r"
        fvals = np.abs(fvals)

    fmax = np.amax(np.abs(fvals))
    if clim is None:
        if cmap=="RdBu":
            clim = (-fmax, fmax)
        else:
            clim = (0, fmax)

    if ax is None:
        fig, ax = plt.subplots(1, constrained_layout=True)

    im = ax.imshow(fvals.T, extent=extent, cmap=cmap, clim=clim, origin='lower')
    if eps_r is not None:
        ealph = min(eps_alpha, 1)
        if (np.amax(eps_r) - np.amin(eps_r) > 1e-10):
            _plot_eps(eps_r.T, cmap=cmap_eps, clim=None, ax=ax, extent=extent,
                        cbar=False, alpha=eps_alpha)

    if cbar==True:
        plt.colorbar(im, ax=ax, shrink=0.8)

    ax.set_xlabel(x_lab + " (um)")
    ax.set_ylabel(y_lab + " (um)")
    ax.set_title(ax_title)

    return im

def _check_monitor(mnt_data, field):
    if field.lower()=='e':
        if mnt_data.E.size==0:
            log_and_raise(
                "Monitor has no stored E-field values.",
                RuntimeError
            )
    elif field.lower()=='h':
        if mnt_data.H.size==0:
            log_and_raise(
                "Monitor has no stored H-field values.",
                RuntimeError
            )
    elif field.lower()=='s':
        if mnt_data.S.size==0:
            log_and_raise(
                "Monitor has no stored S-field values. Run the simulation and "
                "use Simulation.poynting(Monitor) to compute S.",
                RuntimeError
            )     
    else:
        log_and_raise("'field' can be 'E' or 'H'.", ValueError)

def viz_field_2D(self, monitor, sample_ind=0, normal=None, normal_ind=0,
                    field='E', val='re', comp='z', ax=None, cbar=False,
                    clim=None, eps_alpha=0.3):
    """Plot a 2D cross-section of a field stored in a :class:`.Monitor`.
    
    Parameters
    ----------
    monitor : Monitor
        The monitor to be queried.
    sample_ind : int, optional
        Index of the frequency in a :class:`.FreqMonitor` or the time step in 
        a :class:`.TimeMonitor`.
    normal : None, optional
        Axis normal to the 2D plane of plotting. If ``None``, the shortest 
        dimension is taken as the normal.
    normal_ind : int, optional
        Spatial index along the normal dimension, for 3D monitors.
    field : {'E', 'H', 'S'}, optional
        Which field to plot.
    val : {'re', 'im', 'abs', 'int'}, optional
        Plot the real part (default), or the imaginary part, or the 
        absolute value of a field component, or the total field intensity. 
    comp : {'x', 'y', 'z'}, optional
        Component of the field to plot. If ``val`` is ``'int'``, this 
        parameter has no effect.
    ax : Matplotlib axis object, optional
        If None, a new figure is created. 
    cbar : bool, optional
        Add a colorbar to the plot.
    clim : List[float], optional
        Matplotlib color limit to use for plot.
    eps_alpha : float, optional
        If larger than zero, overlay the underlying permittivity 
        distribution, with opacity defined by ``eps_alpha``.
    
    Returns
    -------
    Matplotlib image object
    """

    mnt_data = self._monitor_ids[id(monitor)]
    _check_monitor(mnt_data, field)

    (fvals, eps_r, extent, x_lab, y_lab, ax_title) = _monitor_slice(
                                    self, mnt_data, sample_ind, normal,
                                    normal_ind, field, val, comp, eps_alpha)

    im = _plot_field_2D(fvals, eps_r, extent, x_lab, y_lab, ax_title,
                        val, ax, cbar, clim, eps_alpha)

    return im

def _fmonitors_png(self, folder_path, val='int', comp='z'):
    """
    Export png images of 2D cross-sections for all frequency monitors in a 
    given simulation. For 3D monitors, all 2D slices along the shortest 
    dimension are exported. 
    
    Parameters
    ----------
    folder_path : string
        Path in which the images will be exported.
    val : str, optional
        ``val`` supplied to :meth:`.viz_field_2D`.
    comp : str, optional
        ``comp`` suplied to :meth:`.viz_field_2D`.
    """

    fig = plt.figure(constrained_layout=True)

    for (ip, monitor) in enumerate(self.monitors):

        # Only export frequency monitors
        if isinstance(monitor, TimeMonitor):
            continue 

        mnt_data = self._monitor_ids[id(monitor)]
        mdims = mnt_data.inds_end - mnt_data.inds_beg
        min_dir = np.argmin(mdims)
        normal = xyz_list[min_dir]

        for normal_ind in range(mdims[min_dir]):
            for (find, _) in enumerate(mnt_data.freqs):
                try:
                    ax = fig.add_subplot(111)
                    self.viz_field_2D(monitor, sample_ind=find, normal=normal,
                                normal_ind=normal_ind, val=val, comp=comp,
                                ax=ax, cbar=True)
                    fname = mnt_data.name
                    fname += "_find%d_nind%d.png"%(find, normal_ind)
                    plt.savefig(folder_path+fname)
                    plt.clf()
                except:
                    print("Could not export image for monitor index %d"%ip +
                          ", normal index %d, frequency index %d."%(
                            normal_ind, find))

    plt.close(fig)


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from ..utils import listify, log_and_raise 
from ..monitor import ModeMonitor
from ..source import ModeSource
from .field import _plot_field_2D

def _mode_plane(sim, mode_plane, freq_ind=0, mode_inds=0,
                val='abs', cbar=False, eps_alpha=0.3, clim=None):

    minds = listify(mode_inds)
    grid_list = [(1, 2, 0, 'y', 'z'),
                (0, 2, 1, 'x', 'z'),
                (0, 1, 2, 'x', 'y')]

    (d1, d2, dn, x_lab, y_lab) = grid_list[mode_plane.norm_ind]
    N1, N2 = mode_plane.eps.shape

    if eps_alpha<=0:
        eps_r = None
    else:
        eps_r = mode_plane.eps

    mesh_c1 = mode_plane.mesh[0]
    mesh_c2 = mode_plane.mesh[1]
    extent = [mesh_c1[0], mesh_c1[-1], mesh_c2[0], mesh_c2[-1]]
    aspect = (extent[3]-extent[2])/(extent[1]-extent[0])

    fig, axs = plt.subplots(len(minds), 2, figsize=(8, 4*aspect*len(minds)),
                            constrained_layout=True)
    if len(minds)==1:
        axs = axs.reshape((1, 2))

    for iax, imode in enumerate(minds):
        (E, _) = mode_plane.modes[freq_ind][imode].fields_to_center()
        Ec1 = E[0, :, :]
        Ec2 = E[1, :, :]

        _clim = clim
        if clim is None:
            cmax = np.amax(np.abs(np.vstack((Ec1, Ec2))))
            if val=='abs':
                _clim = (0, cmax)
            else:
                _clim = (-cmax, cmax)

        subtitle = "f=%1.2eTHz, "%(mode_plane.freqs[freq_ind]*1e-12)
        subtitle += "n=%1.2f"%mode_plane.modes[freq_ind][imode].neff
        ax_title = "Mode %d, E%s"%(imode, x_lab) + "\n" + subtitle
        _plot_field_2D(Ec1, eps_r, extent, x_lab, y_lab, ax_title,
                        val=val, ax=axs[iax, 0], cbar=False, clim=_clim,
                        eps_alpha=eps_alpha)
        ax_title = "Mode %d, E%s"%(imode, y_lab) + "\n" + subtitle
        _plot_field_2D(Ec2, eps_r, extent, x_lab, y_lab, ax_title,
                        val=val, ax=axs[iax, 1], cbar=cbar, clim=_clim,
                        eps_alpha=eps_alpha)

    return fig

def viz_modes(self, mode_obj, mode_inds=0, freq_ind=0,
                val='abs', cbar=False, clim=None, eps_alpha=0.3):
    """Plot the field distribution of the 2D eigenmodes of 
    a :class:`.ModeSource` or a :class:`.ModeMonitor` object.
    
    Parameters
    ----------
    mode_obj : ModeSource or ModeMonitor
        An object on which :meth:`.Simulation.compute_modes` can be used.
    mode_inds : array_like, optional
        Mode indexes of the stored modes to be plotted.
    freq_ind : int, optional
        Frequency index of the stored modes to be plotted.
    val : {'re', 'im', 'abs'}, optional
        Plot the real part (default), or the imaginary or absolute value of 
        the field components.
    cbar : bool, optional
        Add a colorbar to the plot.
    clim : List[float], optional
        Matplotlib color limit to use for plot.
    eps_alpha : float, optional
        If larger than zero, overlay the underlying permittivity distribution, 
        with opacity defined by eps_alpha.

    Returns
    -------
    Matplotlib figure object

    Note
    ----
    If the modes have not been computed yet, or ``mode_inds`` exceeds the 
    largest mode index that is stored in ``mode_obj``, 
    :meth:`.Simulation.compute_modes` will be called.
    """

    if isinstance(mode_obj, ModeSource):
        mode_pl = self._source_ids[id(mode_obj)].mode_plane
    elif isinstance(mode_obj, ModeMonitor):
        mode_pl = self._monitor_ids[id(mode_obj)].mode_plane
    else:
        log_and_raise(
            "Input 0 should be an instance of ModeSource or ModePlane.",
            ValueError
        )

    nfreqs = len(mode_pl.modes)

    # Check if frequency index out of bounds
    if nfreqs < freq_ind:
        log_and_raise(
            f"Frequency index {freq_ind} our of bounds for "
            f"stored modes number of frequencies {nfreqs}.",
            ValueError
        )

    # Check if modes have not been computed yet, or if fewer than requested
    if len(mode_pl.modes[freq_ind]) <= np.amax(mode_inds):
        self.compute_modes(mode_obj, np.amax(mode_inds) + 1)

    fig = _mode_plane(self, mode_pl, freq_ind, mode_inds,
                val=val, cbar=cbar, clim=clim, eps_alpha=eps_alpha)

    return fig


import numpy as np

class Mode(object):
    """
    Eigenmode of a 2D cross-section in a 3D simulation, assuming translational 
    symmetry in the third dimesnion.
    """
    def __init__(self, E, H, neff, keff):
        """Construct.
        
        Parameters
        ----------
        E : array_like
            Electric field of the mode.
        H : array_like
            Magnetic field of the mode, must be same shape as ``E```.
        neff : float
            Real part of the effective index.
        keff : float
            Imaginary part of the effective index.
        """
        self.E = E
        self.H = H
        self.neff = neff
        self.keff = keff

        # Effective k-vector
        # TODO: this is taken to be purely real later on and scaled to 
        # self.neff, but generally, it can be complex.
        self.kvector = None

    def fields(self):
        return (self.E, self.H)

    def fields_to_center(self):
        """ Return the E and H fields interpolated at the center of the Yee 
        cells in the 2D cross-section. """

        E = np.copy(self.E)/2
        H = np.copy(self.H)/2
        E[0, :, :] += np.roll(E[0, :, :], -1, 1)
        E[1, :, :] += np.roll(E[1, :, :], -1, 0)
        H[0, :, :] += np.roll(H[0, :, :], -1, 0)
        H[1, :, :] += np.roll(H[1, :, :], -1, 1)

        return (E, H)

    def to_dict(self):
        """ Returns a dictionary with the Mode properties. """
        return {'E': self.E, 
                'H': self.H,
                'neff': self.neff,
                'keff': self.keff
                }

    def fix_efield_phase(self):
        """ Apply a global phase based on the point with the largest E-field 
        value. This value is fixed to be real and positive. For lossless, 
        localized modes, this global phase should then make all of E and H real.
        """

        ind_max = np.argmax(np.abs(self.E))
        Emax = self.E.ravel()[ind_max]
        phi = np.angle(Emax)
        self.E *= np.exp(-1j*phi)
        self.H *= np.exp(-1j*phi)


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from ..constants import xyz_dict, xyz_list
from ..utils import log_and_raise


def _eps_cmap(eps_r, alpha=1, cmap=None, clim=None):
    if cmap == None:
        cmap = "Greys"
    cm = mpl.cm.get_cmap(cmap, 256)
    newmap = cm(np.linspace(0, 1, 256))
    newmap[:, 3] = alpha

    # Make eps = 1 completely transparent
    newmap[0, 3] = 0

    if clim is None:
        clim = (1, np.amax(eps_r) + 1e-6)

    bounds = np.hstack((-4e6, -2e6, 0, np.linspace(clim[0], clim[1], 256)))
    eps_cmap = np.vstack(([0, 0, 1, 1], [1, 0, 0, 1], [1, 1, 1, 0], newmap))
    norm = mpl.colors.BoundaryNorm(boundaries=bounds, ncolors=259)

    return mpl.colors.ListedColormap(eps_cmap), norm, bounds


def _get_inside(objects, mesh):
    """Get a mask defining points inside a list of objects.

    Parameters
    ----------
    objects : list of Structure, Source, or Monitor objects
    mesh : tuple of 3 1D arrays, or None

    Returns
    -------
    mask : np.ndarray
        Array of size (mesh[0].size, mesh[1].size, mesh[2].size) where each
        element is one if inside any of the objects, and zero otherwise.
    """

    Nx, Ny, Nz = [mesh[i].size for i in range(3)]

    mask = np.zeros((Nx, Ny, Nz))
    for obj in objects:
        mtmp = obj._inside(mesh)
        mask[mtmp > 0] = 1
        # print(obj, np.sum(mask))

    return mask


def _plot_eps(
    eps_r,
    cmap=None,
    clim=None,
    ax=None,
    extent=None,
    cbar=False,
    cax=None,
    alpha=1,
):

    if ax is None:
        fig, ax = plt.subplots(1, constrained_layout=True)

    cmplot, norm, bounds = _eps_cmap(eps_r, alpha, cmap, clim)

    im = ax.imshow(
        eps_r,
        interpolation="none",
        norm=norm,
        cmap=cmplot,
        origin="lower",
        extent=extent,
    )

    if cbar:
        if cax is not None:
            plt.colorbar(im, ax=ax, cax=cax, boundaries=bounds[3:])
        else:
            plt.colorbar(im, ax=ax, boundaries=bounds[3:])

    return im


def viz_eps_2D(
    self,
    normal="x",
    position=0.0,
    ax=None,
    cbar=False,
    clim=None,
    source_alpha=0.3,
    monitor_alpha=0.3,
    pml_alpha=0.2,
):
    """Plot the relative permittivity distribution of a 2D cross-section of
    the simulation.

    Parameters
    ----------
    normal : {'x', 'y', 'z'}
        Axis normal to the cross-section plane.
    position : float, optional
        Position offset along the normal axis.
    ax : Matplotlib axis object, optional
        If ``None``, a new figure is created.
    cbar : bool, optional
        Add a colorbar to the plot.
    clim : List[float], optional
        Matplotlib color limit to use for plot.
    source_alpha : float, optional
        If larger than zero, overlay all sources in the simulation,
        with opacity defined by ``source_alpha``.
    monitor_alpha : float, optional
        If larger than zero, overlay all monitors in the simulation,
        with opacity defined by ``monitor_alpha``.
    pml_alpha : float, optional
        If larger than zero, overlay the PML boundaries of the simulation,
        with opacity defined by ``pml_alpha``.

    Returns
    -------
    Matplotlib image object

    Note
    ----
    The plotting is discretized at the center positions of the Yee grid and
    is for illustrative purposes only. In the FDTD computation, the exact
    Yee grid is used and the permittivity values depend on the field
    polarization.
    """

    grid = self.grid

    if ax is None:
        fig, ax = plt.subplots(1, constrained_layout=True)

    # Get normal and cross-section axes indexes
    norm = xyz_dict[normal]
    cross = [0, 1, 2]
    cross.pop(norm)

    # Get centered mesh for permittivity discretization
    mesh = [[], [], []]
    ind = np.nonzero(position < grid.mesh[norm])[0]
    if ind.size == 0:
        log_and_raise(
            "Plane position outside of simulation domain.",
            RuntimeError
        )
    else:
        ind = ind[0]
    mesh[norm] = np.array([grid.mesh[norm][ind]])
    mesh[cross[0]] = grid.mesh[cross[0]]
    mesh[cross[1]] = grid.mesh[cross[1]]

    eps_r = np.squeeze(self._get_eps(mesh))

    # Get mesh for source and monitor discretization
    if ind > 0:
        mesh_sp = [[], [], []]
        mesh_sp[norm] = grid.mesh[norm][ind - 1 : ind + 1]
        mesh_sp[cross[0]] = mesh[cross[0]]
        mesh_sp[cross[1]] = mesh[cross[1]]

    # Set axes xlim and ylim if not provided
    sim_xlim = (grid.mesh[cross[0]][0], grid.mesh[cross[0]][-1])
    sim_ylim = (grid.mesh[cross[1]][0], grid.mesh[cross[1]][-1])

    # Define axes extent, labels and title
    extent = list(sim_xlim) + list(sim_ylim)
    x_lab = xyz_list[cross[0]]
    y_lab = xyz_list[cross[1]]
    ax_tit = x_lab + y_lab
    ax_tit += "-plane at " + xyz_list[norm] + "=%1.2eum" % position

    # Transparent color
    transp = (0, 0, 0, 0)
    # Colors for monitors, sources, pml, and border
    mnt_color = (236 / 255, 203 / 255, 32 / 255)
    src_color = (78 / 255, 145 / 255, 78 / 255)
    pml_color = (229 / 255, 127 / 255, 25 / 255)
    pbc_color = (75 / 255, 0, 130 / 255)

    # # Set axis background color
    # ax.set_facecolor(pml_color)

    # Plot and set axes properties
    im = _plot_eps(eps_r.T, clim=clim, ax=ax, extent=extent, cbar=cbar)
    ax.set_xlabel(x_lab + " (um)")
    ax.set_ylabel(y_lab + " (um)")
    ax.set_title(ax_tit)

    # Plot simulation domain borders depending on boundaries
    npml = self.Npml[[cross[0], cross[1]], :]
    # In order of bottom, right, top, left
    border_pmls = [npml[1, 0], npml[0, 1], npml[1, 1], npml[0, 0]]
    x_border = [extent[0], extent[1], extent[1], extent[0], extent[0]]
    y_border = [extent[2], extent[2], extent[3], extent[3], extent[2]]
    for ib, bpml in enumerate(border_pmls):
        if bpml > 0:
            border_color = pml_color
        else:
            border_color = pbc_color
        ax.plot(
            [x_border[ib], x_border[ib+1]],
            [y_border[ib], y_border[ib+1]],
            color=border_color
        )

    def squeeze_mask(mask, axis):
        inds = [slice(None), slice(None), slice(None)]
        inds[axis] = 1
        return np.squeeze(mask[tuple(inds)])

    if monitor_alpha > 0:
        mnt_alpha = min(monitor_alpha, 1)
        mnt_c = list(mnt_color)
        mnt_c.append(mnt_alpha)
        mnt_mask = squeeze_mask(_get_inside(self.monitors, mesh=mesh_sp), norm)
        mnt_cmap = mpl.colors.ListedColormap(np.array([transp, mnt_c]))
        ax.imshow(
            mnt_mask.T,
            clim=(0, 1),
            cmap=mnt_cmap,
            origin="lower",
            extent=extent,
        )

    if source_alpha > 0:
        src_alpha = min(source_alpha, 1)
        src_c = list(src_color)
        src_c.append(src_alpha)
        src_mask = squeeze_mask(_get_inside(self.sources, mesh=mesh_sp), norm)
        src_cmap = mpl.colors.ListedColormap(np.array([transp, src_c]))
        ax.imshow(
            src_mask.T,
            clim=(0, 1),
            cmap=src_cmap,
            origin="lower",
            extent=extent,
        )

    if pml_alpha > 0:
        pml_alpha = min(pml_alpha, 1)
        pml_c = list(pml_color)
        pml_c.append(pml_alpha)
        pml_mask = np.squeeze(
            np.zeros((mesh[0].size, mesh[1].size, mesh[2].size))
        )
        N1, N2 = pml_mask.shape
        pml_mask[: npml[0, 0], :] = 1
        pml_mask[N1 - npml[0, 1] :, :] = 1
        pml_mask[:, : npml[1, 0]] = 1
        pml_mask[:, N2 - npml[1, 1] :] = 1
        pml_cmap = mpl.colors.ListedColormap(np.array([transp, pml_c]))
        ax.imshow(
            pml_mask.T,
            clim=(0, 1),
            cmap=pml_cmap,
            origin="lower",
            extent=extent,
        )

    return im


def _structure_png(self, folder_path, aspect_ratio=None):
    """
    Export png images of 2D cross-sections of the simulation at the domain
    center in x, y, and z.

    Parameters
    ----------
    folder_path : string
        Path in which the images will be exported.
    aspect_ratio : float or None
        Fix the width/height image ratio. If ``None``, the smallest aspect 
        ratio of the three plots is used in all.
    """

    fig = plt.figure(constrained_layout=True)

    if not aspect_ratio:
        # Take the smallest aspect ratio
        sim_ars = [
            self.size[1] / self.size[2],
            self.size[0] / self.size[2],
            self.size[0] / self.size[1],
        ]
        aspect_ratio = np.min(sim_ars)

    for norm_ind, normal in enumerate(["x", "y", "z"]):
        cinds = [0, 1, 2]
        cinds.pop(norm_ind)
        ax = fig.add_subplot(111)
        self.viz_eps_2D(normal=normal, position=self.center[norm_ind], ax=ax)
        sim_width = self.size[cinds[0]]
        sim_height = self.size[cinds[1]]
        sim_aspect = sim_width / sim_height
        if sim_aspect > aspect_ratio:
            new_height = sim_width / aspect_ratio
            ax.set_ylim(-new_height / 2, new_height / 2)
        elif sim_aspect < aspect_ratio:
            new_width = sim_height * aspect_ratio
            ax.set_xlim(-new_width / 2, new_width / 2)
        fname = "simulation_%scent.png" % normal
        plt.savefig(folder_path + fname, dpi=150)
        plt.clf()

    plt.close(fig)



import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl

from ..constants import EPSILON_0, ETA_0, C_0, MU_0
from .derivatives import compute_derivative_matrices
from .Mode import Mode

def compute_modes(eps_cross, freq, mesh_step, pml_layers, num_modes=1):
    """Solve for the modes of a waveguide cross section.
    
    Parameters
    ----------
    eps_cross : array_like or tuple of array_like
        Either a single 2D array defining the relative permittivity in the 
        cross-section, or three 2D arrays defining the permittivity at the Ex, 
        Ey, and Ez locations of the Yee cell, respectively.
    freq : float
        (Hertz) Frequency at which the eigenmodes are computed.
    mesh_step : list or tuple of float
        (micron) Step size in x, y and z. The mesh step in z is currently 
        unused, but it could be needed if numerical dispersion is to be taken 
        into account.
    pml_layers : list or tuple of int
        Number of pml layers in x and y.
    num_modes : int, optional
        Number of modes to be computed.
    
    Returns
    -------
    list of dict
        A list of all the computed modes. Each entry is a dictionary with the 
        real and imaginary part of the effective index of the waveguide and 
        all the E and H field components.
    """
    omega = 2*np.pi*freq
    k0 = omega / C_0

    dt = 7.281106e-17
    # k0 = np.sin(dt*omega/2)/dt*2/C_0

    if isinstance(eps_cross, np.ndarray):
        eps_xx, eps_yy, eps_zz = [eps_cross]*3
    else:
        eps_xx, eps_yy, eps_zz = eps_cross

    Nx, Ny = eps_xx.shape
    N = eps_xx.size

    matrices = compute_derivative_matrices(omega, (Nx, Ny), tuple(pml_layers), 
                                            tuple(mesh_step[:2]))

    # Normalize by k0 to match the EM possible notation
    Dxf, Dxb, Dyf, Dyb = [mat/k0 for mat in matrices]

    inv_eps_zz = sp.spdiags(1/eps_zz.flatten(), [0], N, N)
    P11 = -Dxf.dot(inv_eps_zz).dot(Dyb)
    P12 = Dxf.dot(inv_eps_zz).dot(Dxb) + sp.eye(N)
    P21 = -Dyf.dot(inv_eps_zz).dot(Dyb) - sp.eye(N)
    P22 = Dyf.dot(inv_eps_zz).dot(Dxb)
    Q11 = -Dxb.dot(Dyf)
    Q12 = Dxb.dot(Dxf) + sp.spdiags(eps_yy.flatten(), [0], N, N)
    Q21 = -Dyb.dot(Dyf) - sp.spdiags(eps_xx.flatten(), [0], N, N)
    Q22 = Dyb.dot(Dxf)

    Pmat = sp.bmat([[P11, P12], [P21, P22]])
    Qmat = sp.bmat([[Q11, Q12], [Q21, Q22]])
    A = Pmat.dot(Qmat)

    n_max = np.sqrt(np.max(eps_cross))
    vals, vecs = solver_eigs(A, num_modes, guess_value=-n_max**2)
    vre, vim = -np.real(vals), -np.imag(vals)

    # Sort by descending real part
    sort_inds = np.argsort(vre)[::-1]
    vre = vre[sort_inds]
    vim = vim[sort_inds]
    vecs = vecs[:, sort_inds]

    # Real and imaginary part of the effective index
    neff_tmp = np.sqrt(vre/2 + np.sqrt(vre**2 + vim**2)/2)
    keff = vim/2/(neff_tmp + 1e-10)

    # Correct formula taking numerical dispersion into account
    # neff = 2/mesh_step[2]*np.arcsin((neff_tmp + 1j*keff)*mesh_step[2]/2)
    neff = neff_tmp

    # Field components from eigenvectors
    Ex = vecs[:N, :]
    Ey = vecs[N:, :]

    # Get the other field components; normalize according to CEM
    Hs = -Qmat.dot(vecs)/(neff + 1j*keff)[np.newaxis, :]/ETA_0
    Hx = Hs[:N, :]
    Hy = Hs[N:, :]

    Hz = Dxf.dot(Ey) - Dyf.dot(Ex)
    Ez = inv_eps_zz.dot((Dxb.dot(Hy) - Dyb.dot(Hx)))

    # Store all the information about the modes.
    modes = []
    for im in range(num_modes):
        E = np.array([Ex[:, im].reshape(Nx, Ny),
                        Ey[:, im].reshape(Nx, Ny),
                        Ez[:, im].reshape(Nx, Ny)])
        H = np.array([Hx[:, im].reshape(Nx, Ny),
                        Hy[:, im].reshape(Nx, Ny),
                        Hz[:, im].reshape(Nx, Ny)])
        modes.append(Mode(E, H, neff[im], keff[im]))

    if vals.size == 0:
        raise RuntimeError("Could not find any eigenmodes for this waveguide")

    return modes

def solver_eigs(A, num_modes, guess_value=1.0):
    """ Find ``num_modes`` eigenmodes of ``A`` cloest to ``guess_value``.

    Parameters
    ----------
    A : scipy.sparse matrix
        Square matrix for diagonalization.
    num_modes : int
        Number of eigenmodes to compute.
    guess_value : float, optional
    """

    values, vectors = spl.eigs(A, k=num_modes, sigma=guess_value)
    return values, vectors


import numpy as np

from .Monitor import TimeMonitor, FreqMonitor, ModeMonitor
from ..constants import int_, float_, complex_, fp_eps, C_0
from ..mode import ModePlane

class MonitorData(object):
    """Class used internally to store data related to a Monitor added in a 
    specific Simulation.
    """

    def __init__(self, monitor):

        # Reference to the monitor that the data corresponds to
        self.monitor = monitor

        # To be set by Simulation
        self.name = None

        # Everything below is set after a run of store_data() or load_fields()
        self.data = False # To be set to True after data is loaded

        # Raw E and H data as returned by the solver
        self._E = np.empty((3, 0, 0, 0, 0), dtype=float_)
        self._H = np.empty((3, 0, 0, 0, 0), dtype=float_)
        # Poynting flux if computed after a solve
        self.S = np.empty((3, 0, 0, 0, 0), dtype=float_)

        # Mesh defining the positions at which the fields are returned
        self.xmesh = np.empty((0, ), dtype=float_)
        self.ymesh = np.empty((0, ), dtype=float_)
        self.zmesh = np.empty((0, ), dtype=float_)
        self.mesh_step = np.empty((3, ), dtype=float_)

        # Mesh in time, for a TimeMonitor
        self.tmesh = np.empty((0, ), dtype=float_)
        self.tind_beg = 0
        self.tind_end = 0
        self.Nt = 0

        # Frequencies, for a FreqMonitor
        self.freqs = np.empty((0, ), dtype=float_)
        if isinstance(monitor, FreqMonitor):
            self.freqs = np.array(monitor.freqs, dtype=float_)

        # Indexes defining the span of the Monitor in the simulation grid in 
        # which it is embedded.
        self.mnt_inds = np.empty((0, 3), dtype=int_)
        self.inds_beg = np.zeros((3, ), dtype=int_)
        self.inds_end = np.zeros((3, ), dtype=int_)

        # Source normalization, which may be set later for FreqMonitors
        self.set_source_norm(None) 

    # def __repr__(self):
    #     if self.data==False:
    #         rep += "Has data: False"
    #     else:
    #         rep += "Has data: xmesh, ymesh, zmesh, freqs, " + ", ".join(
    #                                         [f.upper() for f in self.field])
    #         if hasattr(self, 'S'):
    #             rep += ", S"

    #     rep += "}\n"
    #     return rep

    @property
    def E(self):
        """ (V/m) Electric field, if it was requested by the monitor and 
        after the data has been loaded.
        """
        return self._E * self.source_norm[None, None, None, None, :]

    @property
    def H(self):
        """ (A/m) Magnetic field, if it was requested by the monitor and 
        after the data has been loaded.
        """
        return self._H * self.source_norm[None, None, None, None, :]

    def _set_tmesh(self, tmesh):
        """ Set the time mesh of the monitor. At the end of a simulation run, 
        the field values are syncronized to be on the ``tmesh`` as defined 
        here. This involves recording one extra step in the solver, and 
        averaging the H-fields, which live on a (n + 1/2)*dt time mesh.
        """
 
        # Step to compare to in order to handle t_start = t_stop
        if np.array(tmesh).size < 1:
            dt = 1e-20
        else:
            dt = tmesh[1] - tmesh[0]

        # If t_stop is None, record until the end
        t_stop = self.monitor.t_stop
        if self.monitor.t_stop is None:
            t_stop = tmesh[-1]
            self.tind_end = tmesh.size
        else:
            tend = np.nonzero(tmesh <= self.monitor.t_stop)[0]
            if tend.size > 0:
                self.tind_end = tend[-1] + 1
            else:
                self.tind_end = 0

        # If equal (within dt), record one time step
        if np.abs(self.monitor.t_start - t_stop) < dt:
            self.tind_beg = np.max([self.tind_end - 1, 0])
        else:
            tbeg = np.nonzero(tmesh[0:self.tind_end] >= 
                                self.monitor.t_start)[0]
            if tbeg.size > 0:
                self.tind_beg = tbeg[0]
            else:
                self.tind_beg = self.tind_end

        self.tmesh = tmesh[self.tind_beg:self.tind_end]
        self.Nt = self.tind_end - self.tind_beg

    def _set_mode_plane(self):
        self.norm_ind = self.monitor.norm_ind
        self.cross_inds = [0, 1, 2]
        self.cross_inds.pop(self.norm_ind)
        self.mode_plane = ModePlane(self.monitor.span, self.norm_ind)

    def _store_data(self, mnt_data, mnt_inds, field):
        """ Store the raw monitor values returned by the solver.
        
        Parameters
        ----------
        mnt_data : np.ndarray
            An array of shape ``(Np, 3, Nsample)``, where ``Np`` is the total 
            number of monitor points, and ``Nsample`` is either the number of 
            time steps in a ``TimeMonitor``, or the number of requested 
            frequencies in a ``FreqMonitor``.
        mnt_inds : np.ndarray
            An array of shape ``(Np, 3)`` giving the x, y, and z index for 
            each point in the simulation grid.

        Note
        ----
        ``MonitorData.E`` and ``MonitorData.H`` are stored in the format 
        ``(3, indx, indy, indz, samp_ind)``, where ``samp_ind`` is either a 
        time or a frequency index.
        """

        self.mnt_inds = mnt_inds
        self.inds_beg = np.amin(mnt_inds, axis = 0)
        self.inds_end = np.amax(mnt_inds, axis = 0) + 1
        mnt_dims = self.inds_end - self.inds_beg
        if field.lower()=='e':
            self._E = np.zeros((3, mnt_dims[0], mnt_dims[1], mnt_dims[2], 
                        mnt_data.shape[2]), dtype=mnt_data.dtype)
            for ipol in range(3):
                self._E[ipol, mnt_inds[:, 0]-self.inds_beg[0],
                    mnt_inds[:, 1]-self.inds_beg[1],
                    mnt_inds[:, 2]-self.inds_beg[2], :] = mnt_data[:, ipol, :]
        elif field.lower()=='h':
            self._H = np.zeros((3, mnt_dims[0], mnt_dims[1], mnt_dims[2], 
                        mnt_data.shape[2]), dtype=mnt_data.dtype)
            for ipol in range(3):
                self._H[ipol, mnt_inds[:, 0]-self.inds_beg[0],
                    mnt_inds[:, 1]-self.inds_beg[1],
                    mnt_inds[:, 2]-self.inds_beg[2], :] = mnt_data[:, ipol, :]
        self.data = True

    def _load_fields(self, inds_beg, inds_end, E, H, symmetries, Nxyz):
        """ Load the fields returned by the solver. This also applies any 
        symmetries that were present in the simulation.
        """

        # By default just store index and field values, unless changed below
        self.inds_beg = inds_beg
        self.inds_end = inds_end
        self._E = E
        self._H = H

        # Auxiliary variable for slicing along a given axis
        slices = (slice(None), slice(None), slice(None),
                        slice(None), slice(None))

        for dim, sym in enumerate(symmetries):

            # Auxiliary variable for symmetry eigenvalue along current axis
            svals = np.ones((3, 1, 1, 1, 1))
            svals[dim] = -1

            if sym==-1 and inds_beg[dim]==0:
                self.inds_beg[dim] = Nxyz[dim]//2 - inds_end[dim]
                self.inds_end[dim] = Nxyz[dim]//2 + inds_end[dim]
                sl = list(slices)
                sl[dim+1] = slice(-1, None, -1)
                if E.size > 0:
                    self._E = np.concatenate((-svals*self._E[tuple(sl)],
                                self._E), axis=dim+1)
                if H.size > 0:
                    self._H = np.concatenate((svals*self._H[tuple(sl)],
                                self._H), axis=dim+1)

            if sym==1 and inds_beg[dim]==0:
                ibeg = 1
                iend = min(inds_end[dim], Nxyz[dim]//2+1)
                self.inds_beg[dim] = Nxyz[dim]//2 - (iend - ibeg)
                self.inds_end[dim] = Nxyz[dim]//2 + (iend - ibeg)
                sl1 = list(slices)
                sl1[dim+1] = slice(iend-1, ibeg-1, -1)
                sl2 = list(slices)
                sl2[dim+1] = slice(ibeg, iend)  
                if E.size > 0:
                    self._E = np.concatenate((svals*self._E[tuple(sl1)],
                                self._E[tuple(sl2)]), axis=dim+1)
                if H.size > 0:
                    self._H = np.concatenate((-svals*self._H[tuple(sl1)],
                                self._H[tuple(sl2)]), axis=dim+1)
        self.data = True

    def set_source_norm(self, src_data):
        """Normalize the stored fields.
        
        Parameters
        ----------
        src_data : src_data or None
            A :class:`.SourceData` object from whose spectrum the fields are 
            normalized.
        """

        if isinstance(self.monitor, TimeMonitor):
            self.source_norm = np.ones((1, ), dtype=float_)
        elif src_data is None:
            self.source_norm = np.ones((len(self.freqs), ), dtype=float_)
        else:
            spectrum = src_data._get_spectrum(self.freqs)
            self.source_norm = np.array(1/np.abs(spectrum))


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

version = '0.2.0.11'
setup(
    name='tidy3dclient',
    version=version,
    description='A Python API for Tidy3D FDTD Solver',
    author='FlexCompute, Inc.',
    author_email='lei@flexcompute.com',
    packages=find_packages(),
    install_requires=['aws-requests-auth', 'bcrypt'] + requirements,
    dependency_links=['http://github.com/flexcompute/warrant/tarball/master#egg=warrant-0.6.4'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)



import numpy as np

from .misc import listify
from .log import log_and_raise

def inside_box_inds(span, mesh, include_zero_size=False):
    """Elementwise indicator function showing which points in a Box ``span``
    are inside a grid defined by ``mesh``.

    Parameters
    ----------
    span : np.ndarray of shape (3, 2)
        Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the box.
    mesh : tuple
        3-tuple defining the xgrid, ygrid and zgrid.
    include_zero_size : bool, optional
        If True, a span of zero (max - min = 0) in any direction will be taken
        as one pixel. If False, the mask will be all empty.

    Returns
    -------
    indsx, indsy, indsz: tuple
        Tuples defining the (start, stop) index of the mesh inside the box.
    """

    # Check if min and max values are in order
    for dim in range(3):
        if span[dim, 1] < span[dim, 0]:
            log_and_raise(
                "Incorrect object span (max value smaller than " "min value).",
                ValueError
            )

    # Initialize indexes for no part of the mesh inside the box
    indsx, indsy, indsz = ((0, 0), (0, 0), (0, 0))

    # Do some checks that should return no mesh inside the box
    for dim in range(3):
        # Check if any of the mesh arrays has zero lengt
        if mesh[dim].size == 0:
            return indsx, indsy, indsz
        # Check if box is completely outside mesh
        if (span[dim, 0] > mesh[dim][-1]) or (span[dim, 1] < mesh[dim][0]):
            return indsx, indsy, indsz
        # Check if there's a strictly zero span and it was not requested.
        if (span[dim, 1] - span[dim, 0] == 0) and (not include_zero_size):
            return indsx, indsy, indsz

    indx = np.nonzero((span[0, 0] < mesh[0]) * (mesh[0] < span[0, 1]))[0]
    indy = np.nonzero((span[1, 0] < mesh[1]) * (mesh[1] < span[1, 1]))[0]
    indz = np.nonzero((span[2, 0] < mesh[2]) * (mesh[2] < span[2, 1]))[0]

    # If there's still a zero span and it was not requested, we're done.
    if (indx.size == 0 or indy.size == 0 or indz.size == 0) and (
        not include_zero_size
    ):
        return indsx, indsy, indsz

    # Otherwise, we build the mask. For zero span, we just take the first
    # element of the mesh that is larger than the requested min value.
    if indx.size == 0:
        indx1 = np.nonzero(span[0, 0] <= mesh[0])[0][0]
        indsx = (indx1, indx1 + 1)
    else:
        indsx = (indx[0], indx[-1] + 1)
    if indy.size == 0:
        indy1 = np.nonzero(span[1, 0] <= mesh[1])[0][0]
        indsy = (indy1, indy1 + 1)
    else:
        indsy = (indy[0], indy[-1] + 1)
    if indz.size == 0:
        indz1 = np.nonzero(span[2, 0] <= mesh[2])[0][0]
        indsz = (indz1, indz1 + 1)
    else:
        indsz = (indz[0], indz[-1] + 1)

    return (indsx, indsy, indsz)


def inside_box(span, mesh, include_zero_size=False):
    """Elementwise indicator function showing which points in the ``span`` are
    inside a grid defined by ``mesh``.

    Parameters
    ----------
    span : np.ndarray of shape (3, 2)
        Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the box.
    mesh : tuple
        3-tuple defining the xgrid, ygrid and zgrid.
    include_zero_size : bool, optional
        If True, a span of zero (max - min = 0) in any direction will be taken
        as one pixel. If False, the mask will be all empty.

    Returns
    -------
    mask : np.ndarray
        A 3D array of shape (mesh[0].size, mesh[1].size, mesh[2].size)
        that is 1 inside the box and 0 outside.
    """

    # Initialize empty mask
    mask = np.zeros((mesh[0].size, mesh[1].size, mesh[2].size))

    indsx, indsy, indsz = inside_box_inds(span, mesh, include_zero_size)
    mask[indsx[0] : indsx[1], indsy[0] : indsy[1], indsz[0] : indsz[1]] = 1.0

    return mask


def intersect_box(span1, span2):
    """Return a span of a box that is the intersection between two spans."""
    span = np.zeros((3, 2))
    for d in range(3):
        span[d, 0] = max(span1[d, 0], span2[d, 0])
        span[d, 1] = min(span1[d, 1], span2[d, 1])

    return span


def cs2span(center, size):
    """Get shape (3, 2) span from center and size, each (3, ) arrays."""
    return np.vstack(
        (
            np.array(center) - np.array(size) / 2,
            np.array(center) + np.array(size) / 2,
        )
    ).T


def span2cs(span):
    """Get center, size: each arrays of shape (3,), from a shape (3, 2) array
    span.
    """
    center = np.array([(span[d, 1] + span[d, 0]) / 2 for d in range(3)])
    size = np.array([(span[d, 1] - span[d, 0]) for d in range(3)])
    return center, size


def axes_handed(axes):
    """Return +1 if the axes indexes in the list ``axes`` form a right-handed
    coordinate system, and -1 if they form a left-handed one.
    """

    if listify(axes) in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
        return 1
    elif listify(axes) in [[0, 2, 1], [1, 0, 2], [2, 1, 0]]:
        return -1
    else:
        log_and_raise("Unrecognized list of axes indexes.", ValueError)



import math
from .material import Medium
from .dispersion import DispersionModel, Sellmeier, Lorentz
from .constants import C_0, HBAR


class cSi(Medium):
    """Crystalline silicon at 26 deg C.

    Refs:
    * A. Deinega, I. Valuev, B. Potapkin, and Y. Lozovik, Minimizing light
      reflection from dielectric textured surfaces,
      J. Optical Society of America A, 28, 770-77 (2011).
    * M. A. Green and M. Keevers, Optical properties of intrinsic silicon
      at 300 K, Progress in Photovoltaics, 3, 189-92 (1995).
    * C. D. Salzberg and J. J. Villa. Infrared Refractive Indexes of Silicon,
      Germanium and Modified Selenium Glass,
      J. Opt. Soc. Am., 47, 244-246 (1957).
    * B. Tatian. Fitting refractive-index data with the Sellmeier dispersion
      formula, Appl. Opt. 23, 4477-4485 (1984).
    """

    def __init__(self, variant=None):
        super().__init__(name="cSi")
        self.eps = 1
        self.sigma = 0

        if variant is None:
            variant = "Deinega2011"

        if "SalzbergVilla1957" == variant:
            self.dispmod = Sellmeier(
                [
                    (10.6684293, 0.301516485 ** 2),
                    (0.0030434748, 1.13475115 ** 2),
                    (1.54133408, 1104 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 11.00, C_0 / 1.36)
        elif "Deinega2011" == variant:
            self.dispmod = Lorentz(
                self.eps,
                [
                    (8.000, 3.64 * C_0, 0),
                    (2.850, 2.76 * C_0, 0.063 * C_0),
                    (-0.107, 1.73 * C_0, 2.5 * C_0),
                ],
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.0, C_0 / 0.4)


class aSi(Medium):
    """Amorphous silicon

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="aSi")
        self.eps = 3.109
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(17.68 - self.eps, 3.93 * eV_to_Hz, 0.5 * 1.92 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6 * eV_to_Hz)


class AlAs(Medium):
    """AlAs

    Refs:
    * R.E. Fern and A. Onton, J. Applied Physics, 42, 3499-500 (1971)
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="AlAs")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(8.27 - self.eps, 4.519 * eV_to_Hz, 0.5 * 0.378 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0 * eV_to_Hz, 3 * eV_to_Hz)
        elif "FernOnton1971" == variant:
            self.dispmod = Sellmeier(
                [
                    (6.0840, 0.2822 ** 2),
                    (1.900, 27.62 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 2.2, C_0 / 0.56)


class AlGaN(Medium):
    """AlGaN

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="AlGaN")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(4.6 - self.eps, 7.22 * eV_to_Hz, 0.5 * 0.127 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 4 * eV_to_Hz)


class AlN(Medium):
    """AlN

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="AlN")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(4.306 - self.eps, 8.916 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 4.75 * eV_to_Hz)


class Al2O3(Medium):
    """Alumina

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Al2O3")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.52 - self.eps, 12.218 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 6 * eV_to_Hz)


class AlxOy(Medium):
    """AlxOy

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="AlxOy")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps,
                [(3.171 - self.eps, 12.866 * eV_to_Hz, 0.5 * 0.861 * eV_to_Hz)],
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 6 * eV_to_Hz)


class Aminoacid(Medium):
    """Amino acid

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Aminoacid")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(1.486 - self.eps, 14.822 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 5 * eV_to_Hz)


class CaF2(Medium):
    """CaF2

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="CaF2")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.036 - self.eps, 15.64 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 4.75 * eV_to_Hz)


class GeOx(Medium):
    """GeOx

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="GeOx")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps,
                [(2.645 - self.eps, 16.224 * eV_to_Hz, 0.5 * 0.463 * eV_to_Hz)],
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 4 * eV_to_Hz)


class H2O(Medium):
    """Water

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="H2O")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(1.687 - self.eps, 11.38 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6 * eV_to_Hz)


class HfO2(Medium):
    """HfO2

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="HfO2")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.9 - self.eps, 9.4 * eV_to_Hz, 0.5 * 3.0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6 * eV_to_Hz)


class HMDS(Medium):
    """HMDS

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="HMDS")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.1 - self.eps, 12.0 * eV_to_Hz, 0.5 * 0.5 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6.5 * eV_to_Hz)


class ITO(Medium):
    """Indium Tin Oxide

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="ITO")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.5 - self.eps, 6.8 * eV_to_Hz, 0.5 * 0.637 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6 * eV_to_Hz)


class MgF2(Medium):
    """MgF2

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="MgF2")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(1.899 - self.eps, 16.691 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.8 * eV_to_Hz, 3.8 * eV_to_Hz)


class MgO(Medium):
    """MgO

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="MgO")
        self.eps = 11.232
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.599 - self.eps, 1.0 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 5.5 * eV_to_Hz)


class PEI(Medium):
    """PEI

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PEI")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.09 - self.eps, 12.0 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 4.75 * eV_to_Hz)


class PEN(Medium):
    """PEN

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PEN")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.466 - self.eps, 4.595 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 3.2 * eV_to_Hz)


class PET(Medium):
    """PET

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PET")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.2 - self.eps, 7.0 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = None


class PMMA(Medium):
    """PMMA

    Refs:
    * N. Sultanova, S. Kasarova and I. Nikolov.
      Dispersion properties of optical polymers,
      Acta Physica Polonica A 116, 585-587 (2009)
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PMMA")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Sultanova2009"

        if "Horiba" == variant:
            eV_to_Hz = 0.5 / (math.pi * HBAR)
            self.dispmod = Lorentz(
                self.eps, [(2.17 - self.eps, 11.427 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 4.55 * eV_to_Hz)
        elif "Sultanova2009" == variant:
            self.dispmod = Sellmeier(
                [
                    (1.1819, 0.011313),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.052, C_0 / 0.4368)


class Polycarbonate(Medium):
    """Polycarbonate

    Refs:
    * N. Sultanova, S. Kasarova and I. Nikolov.
      Dispersion properties of optical polymers,
      Acta Physica Polonica A 116, 585-587 (2009)
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Polycarbonate")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Sultanova2009"

        if "Horiba" == variant:
            eV_to_Hz = 0.5 / (math.pi * HBAR)
            self.dispmod = Lorentz(
                self.eps, [(2.504 - self.eps, 12.006 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 4 * eV_to_Hz)
        elif "Sultanova2009" == variant:
            self.dispmod = Sellmeier(
                [
                    (1.4182, 0.021304),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.052, C_0 / 0.4368)


class Polystyrene(Medium):
    """Polystyrene

    Refs:
    * N. Sultanova, S. Kasarova and I. Nikolov.
      Dispersion properties of optical polymers,
      Acta Physica Polonica A 116, 585-587 (2009)
    """

    def __init__(self, variant=None):
        super().__init__(name="Polystyrene")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Sultanova2009"

        if "Sultanova2009" == variant:
            self.dispmod = Sellmeier(
                [
                    (1.4435, 0.020216),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.052, C_0 / 0.4368)


class Cellulose(Medium):
    """Cellulose

    Refs:
    * N. Sultanova, S. Kasarova and I. Nikolov.
      Dispersion properties of optical polymers,
      Acta Physica Polonica A 116, 585-587 (2009)
    """

    def __init__(self, variant=None):
        super().__init__(name="Cellulose")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Sultanova2009"

        if "Sultanova2009" == variant:
            self.dispmod = Sellmeier(
                [
                    (1.124, 0.011087),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.052, C_0 / 0.4368)


class pSi(Medium):
    """pSi

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="pSi")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(12.0 - self.eps, 4.0 * eV_to_Hz, 0.5 * 0.5 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 5 * eV_to_Hz)


class PTFE(Medium):
    """PTFE

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PTFE")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(1.7 - self.eps, 16.481 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 6.5 * eV_to_Hz)


class PVC(Medium):
    """PVC

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="PVC")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.304 - self.eps, 12.211 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 4.75 * eV_to_Hz)


class Sapphire(Medium):
    """Sapphire

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Sapphire")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.09 - self.eps, 13.259 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 5.5 * eV_to_Hz)


class SiC(Medium):
    """SiC

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="SiC")
        self.eps = 3.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(6.8 - self.eps, 8.0 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 4 * eV_to_Hz)


class SiN(Medium):
    """SiN

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="SiN")
        self.eps = 2.32
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.585 - self.eps, 6.495 * eV_to_Hz, 0.5 * 0.398 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 6 * eV_to_Hz)


class Si3N4(Medium):
    """Si3N4

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    * H. R. Philipp. Optical properties of silicon nitride, J. Electrochim. Soc. 120, 295-300 (1973)
    * T. Baak. Silicon oxynitride; a material for GRIN optics, Appl. Optics 21, 1069-1072 (1982)
    * K. Luke, Y. Okawachi, M. R. E. Lamont, A. L. Gaeta, M. Lipson.
      Broadband mid-infrared frequency comb generation in a Si3N4 microresonator,
      Opt. Lett. 40, 4823-4826 (2015)
    """

    def __init__(self, variant=None):
        super().__init__(name="Si3N4")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(5.377 - self.eps, 3.186 * eV_to_Hz, 0.5 * 1.787 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 5.5 * eV_to_Hz)
        elif "Philipp1973" == variant:
            self.dispmod = Sellmeier(
                [
                    (2.8939, 0.13967 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 1.24, C_0 / 0.207)
        elif "Luke2015" == variant:
            self.dispmod = Sellmeier(
                [
                    (3.0249, 0.1353406 ** 2),
                    (40314, 1239.842 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 5.504, C_0 / 0.31)


class SiO2(Medium):
    """SiO2

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="SiO2")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.12 - self.eps, 12.0 * eV_to_Hz, 0.5 * 0.1 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.7 * eV_to_Hz, 5 * eV_to_Hz)


class SiON(Medium):
    """SiON

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="SiON")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.342 - self.eps, 10.868 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 3 * eV_to_Hz)


class Ta2O5(Medium):
    """Ta2O5

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Ta2O5")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(4.133 - self.eps, 7.947 * eV_to_Hz, 0.5 * 0.814 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.75 * eV_to_Hz, 4 * eV_to_Hz)


class TiOx(Medium):
    """TiOx

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="TiOx")
        self.eps = 0.29
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.82 - self.eps, 6.5 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (0.6 * eV_to_Hz, 3 * eV_to_Hz)


class Y2O3(Medium):
    """Y2O3

    Refs:
    * Y. Nigara. Measurement of the optical constants of yttrium oxide,
      Jpn. J. Appl. Phys. 7, 404-408 (1968)
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="Y2O3")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Nigara1968"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(2.715 - self.eps, 9.093 * eV_to_Hz, 0.5 * 0 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.55 * eV_to_Hz, 4 * eV_to_Hz)
        elif "Nigara1968" == variant:
            self.dispmod = Sellmeier(
                [
                    (2.578, 0.1387 ** 2),
                    (3.935, 22.936 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 9.6, C_0 / 0.25)


class ZrO2(Medium):
    """ZrO2

    Refs:
    * Horiba Technical Note 08: Lorentz Dispersion Model
      http://www.horiba.com/fileadmin/uploads/Scientific/Downloads/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf
    """

    def __init__(self, variant=None):
        super().__init__(name="ZrO2")
        self.eps = 1.0
        self.sigma = 0

        if variant is None:
            variant = "Horiba"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Horiba" == variant:
            self.dispmod = Lorentz(
                self.eps, [(3.829 - self.eps, 9.523 * eV_to_Hz, 0.5 * 0.128 * eV_to_Hz)]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (1.5 * eV_to_Hz, 3 * eV_to_Hz)


class BK7(Medium):
    """N-BK7 borosilicate glass"""

    def __init__(self, variant=None):
        super().__init__(name="BK7")
        self.eps = 1
        self.sigma = 0

        if variant is None:
            variant = "Zemax"

        if "Zemax" == variant:
            self.dispmod = Sellmeier(
                [
                    (1.03961212, 0.00600069867),
                    (0.231792344, 0.0200179144),
                    (1.01046945, 103.560653),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 2.5, C_0 / 0.3)


class FusedSilica(Medium):
    """Fused silica

    Refs:
    * I. H. Malitson. Interspecimen comparison of the refractive index of
      fused silica, J. Opt. Soc. Am. 55, 1205-1208 (1965)
    * C. Z. Tan. Determination of refractive index of silica glass for
      infrared wavelengths by IR spectroscopy,
      J. Non-Cryst. Solids 223, 158-163 (1998)
    """

    def __init__(self, variant=None):
        super().__init__(name="FusedSilica")
        self.eps = 1
        self.sigma = 0

        if variant is None:
            variant = "Zemax"

        if "Zemax" == variant:
            self.dispmod = Sellmeier(
                [
                    (0.6961663, 0.0684043 ** 2),
                    (0.4079426, 0.1162414 ** 2),
                    (0.8974794, 9.896161 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 6.7, C_0 / 0.21)


class GaAs(Medium):
    """Gallium arsenide

    Refs:
    * T. Skauli, P. S. Kuo, K. L. Vodopyanov, T. J. Pinguet, O. Levi,
      L. A. Eyres, J. S. Harris, M. M. Fejer, B. Gerard, L. Becouarn,
      and E. Lallier. Improved dispersion relations for GaAs and
      applications to nonlinear optics, J. Appl. Phys., 94, 6447-6455 (2003)
    """

    def __init__(self, variant=None):
        super().__init__(name="GaAs")
        self.eps = 5.372514
        self.sigma = 0

        if variant is None:
            variant = "Skauli2003"

        if "Skauli2003" == variant:
            self.dispmod = Sellmeier(
                [
                    (5.466742, 0.4431307 ** 2),
                    (0.02429960, 0.8746453 ** 2),
                    (1.957522, 36.9166 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 17, C_0 / 0.97)


class Ag(Medium):
    """Silver.

    Refs:
    * A. D. Rakic et al., Applied Optics, 37, 5271-5283 (1998)
    """

    def __init__(self, variant=None):
        super().__init__(name="Ag")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (-2.502e-2 - 8.626e-3j, 5.987e-1 + 4.195e3j),
                    (-2.021e-1 - 9.407e-1j, -2.211e-1 + 2.680e-1j),
                    (-1.467e1 - 1.338e0j, -4.240e0 + 7.324e2j),
                    (-2.997e-1 - 4.034e0j, 6.391e-1 - 7.186e-2j),
                    (-1.896e0 - 4.808e0j, 1.806e0 + 4.563e0j),
                    (-9.396e0 - 6.477e0j, 1.443e0 - 8.219e1j),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.1 / h, 5 / h)


class Au(Medium):
    """Gold

    Refs:
    * P. B. Johnson and R. W. Christy. Optical constants of the noble metals, Phys. Rev. B 6, 4370-4379 (1972)
    """

    def __init__(self, variant=None):
        super().__init__(name="Au")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.026955750077564202 - 0.0988664675659672j),
                        (0.447862536287849 + 359.3693488468651j),
                    ),
                    (
                        (-0.0938891003729386 - 0.4454781163574038j),
                        (-0.7634247243395327 + 3.3378646350288284j),
                    ),
                    (
                        (-0.10241859347785529 - 0.5861784202506137j),
                        (-0.5730996408150796 + 1.5750521669295217j),
                    ),
                    (
                        (-0.5859182121732974 - 2.8095998031643212j),
                        (1.1615305933022986 + 1.529923046941579j),
                    ),
                    (
                        (-2.3648208922948224 - 3.9506663577511176j),
                        (0.4781454204246356 + 10.004764935068316j),
                    ),
                    (
                        (-7.3871945816024915 - 16.048308950027597j),
                        (-3.046940555919119 + 20.73681036538383j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.64 / h, 6.6 / h)


class Cu(Medium):
    """Copper

    Refs:
    * P. B. Johnson and R. W. Christy. Optical constants of the noble metals, Phys. Rev. B 6, 4370-4379 (1972)
    """

    def __init__(self, variant=None):
        super().__init__(name="Cu")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.01754034345120948 - 0.09123699694321295j),
                        (1.0330679957113091 + 356.1912618343321j),
                    ),
                    (
                        (-0.24469644728696058 - 0.16210178020229618j),
                        (-2.115558633152502 + 44.859580087846396j),
                    ),
                    (
                        (-0.4803840208384017 - 2.427821690986015j),
                        (1.3001522133543062 + 2.023013643197286j),
                    ),
                    (
                        (-2.094057267645235 - 4.0383221076029505j),
                        (3.349640664807941 + 8.362216588586927j),
                    ),
                    (
                        (-0.026387014517946214 - 19.200499737531043j),
                        (0.8222623812428774 + 5.49248564124662j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.64 / h, 6.6 / h)


class Al(Medium):
    """Aluminum

    Refs:
    * A. D. Rakic. Algorithm for the determination of intrinsic optical
      constants of metal films: application to aluminum,
      Appl. Opt. 34, 4755-4767 (1995)
    """

    def __init__(self, variant=None):
        super().__init__(name="Al")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.02543000644123838 - 0.03177449689697393j),
                        (2.655977979367176 + 1866.6744251245625j),
                    ),
                    (
                        (-0.0009040206995601725 + 0j),
                        (50.22383221124015 + 148.23535689736036j),
                    ),
                    (
                        (-7.083800421248227 - 0.5265552916184757j),
                        (-10.063691398117694 + 31.243557348153914j),
                    ),
                    (
                        (-0.11804263462150097 - 0.16034526808256572j),
                        (-30.44469672627293 + 50.702553768652265j),
                    ),
                    (
                        (-6.701254199352449 - 3.6481762896415066j),
                        (-11.175149154124156 - 9.307675442873656j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.1 / h, 10 / h)


class Be(Medium):
    """Beryllium

    Refs:
    * A. D. Rakic. Algorithm for the determination of intrinsic optical
      constants of metal films: application to aluminum,
      Appl. Opt. 34, 4755-4767 (1995)
    """

    def __init__(self, variant=None):
        super().__init__(name="Be")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.01467559642300368 - 0.011921713664657874j),
                        (11.102858393689859 + 997.3989259045288j),
                    ),
                    (
                        (-0.007192297005355153 + 0j),
                        (9.17076421804364 + 29779.653005475455j),
                    ),
                    (
                        (-0.1842186369882607 - 0.008345826991602062j),
                        (-12.613037097613995 - 289.8708706194137j),
                    ),
                    (
                        (-1.60693994861672 + 0j),
                        (24.140400506941724 + 84.04918080253373j),
                    ),
                    (
                        (-6.967518750932778 - 6.432541981234564j),
                        (-70.75228325846311 + 4.991406069340526j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.1 / h, 10 / h)


class Cr(Medium):
    """Chromium

    Refs:
    * A. D. Rakic. Algorithm for the determination of intrinsic optical
      constants of metal films: application to aluminum,
      Appl. Opt. 34, 4755-4767 (1995)
    """

    def __init__(self, variant=None):
        super().__init__(name="Cr")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.016744051287845056 - 0.030599362407448455j),
                        (1.5750990207513889 + 236.7100840009308j),
                    ),
                    (
                        (-0.0006926119317860557 + 0j),
                        (13.079639909964918 + 9.498414578099634j),
                    ),
                    (
                        (-0.38748596542834807 - 0.0005273528816316377j),
                        (-14.021929013546274 + 7.538522210064633j),
                    ),
                    (
                        (-1.4809291078222508 - 1.069387688480274j),
                        (0.008744427110544486 + 70.19584519933676j),
                    ),
                    (
                        (-2.181588953284235 - 9.966499320360377j),
                        (-3.262279485359157 - 2.873566458391711j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.1 / h, 10 / h)


class Ni(Medium):
    """Nickel

    Refs:
    * P. B. Johnson and R. W. Christy. Optical constants of the noble metals, Phys. Rev. B 6, 4370-4379 (1972)
    """

    def __init__(self, variant=None):
        super().__init__(name="Ni")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-8.566496800121943e-08 - 0.09838278383411603j),
                        (0.049200052758942875 + 198.13631259069095j),
                    ),
                    (
                        (-0.018141304521375 - 0.10893544701556669j),
                        (5.318590644585675 - 12.472022072681165j),
                    ),
                    (
                        (-0.14928684097140146 - 0.0002279992692829006j),
                        (8.152053607665009 - 8.728651907340463j),
                    ),
                    (
                        (-0.645702819628624 - 0.6004225851474665j),
                        (-0.5915912865455546 + 15.891474422519119j),
                    ),
                    (
                        (-3.0851746199407315 - 5.908453169642721j),
                        (-3.8492168295859273 + 5.786796038682027j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.64 / h, 6.6 / h)


class Pd(Medium):
    """Palladium

    Refs:
    * P. B. Johnson and R. W. Christy. Optical constants of the noble metals, Phys. Rev. B 6, 4370-4379 (1972)
    """

    def __init__(self, variant=None):
        super().__init__(name="Pd")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.018395445268754196 - 0.05793104383593875j),
                        (-0.07689288540916725 + 291.6738427964993j),
                    ),
                    (
                        (-0.027922168600007236 + 0j),
                        (13.774347280659224 - 152.8766207180748j),
                    ),
                    (
                        (-0.760968579570245 - 0.30266586405836354j),
                        (-1.4518130571271635 + 38.71729641003975j),
                    ),
                    (
                        (-0.011091203757874 - 0.01312856138430063j),
                        (-1.4773982002493897 + 147.43877687698034j),
                    ),
                    (
                        (-6.690929831759696 - 4.077751585426895j),
                        (-5.714726349367318 - 1.64330224870603j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (0.64 / h, 6.6 / h)


class Pt(Medium):
    """Platinum

    Refs:
    * W. S. M. Werner, K. Glantschnig, C. Ambrosch-Draxl.
      Optical constants and inelastic electron-scattering data for 17
      elemental metals, J. Phys Chem Ref. Data 38, 1013-1092 (2009)
    """

    def __init__(self, variant=None):
        super().__init__(name="Pt")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.06695203438147745 - 0.14639101632437196j),
                        (3.1173416750016854 + 470.3702792275117j),
                    ),
                    (
                        (-0.05139078154733537 - 0.0398067193616464j),
                        (3.5905379829178328 + 290.51219463657986j),
                    ),
                    (
                        (-4.2702391463452 - 0.1023450079932173j),
                        (3.5169977232897742 + 136.11580292584603j),
                    ),
                    (
                        (-1.5016435398388222 - 0.042541250379806j),
                        (-1.2707409899595357 - 317.59660467815365j),
                    ),
                    (
                        (-6.560611329539 - 2.6604757095535057j),
                        (-0.3302567102379078 + 45.30726384867861j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (C_0 / 0.1, C_0 / 2.48)


class Ti(Medium):
    """Titanium

    Refs:
    * W. S. M. Werner, K. Glantschnig, C. Ambrosch-Draxl.
      Optical constants and inelastic electron-scattering data for 17
      elemental metals, J. Phys Chem Ref. Data 38, 1013-1092 (2009)
    """

    def __init__(self, variant=None):
        super().__init__(name="Ti")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-5.0904093494224216e-05 - 0.13324128147817416j),
                        (-9.262460287562153 + 357.0413551173238j),
                    ),
                    (-0.021305639299682624j, (20.1441560339523 + 13.91885088268019j)),
                    (
                        (-0.05721949014920069 - 0.041292872452438255j),
                        (-5.546364792564378 + 170.84002769123097j),
                    ),
                    (
                        (-3.5847638867645655 - 3.8101711951435884j),
                        (6.0646767851686985 + 5.5853872618611105j),
                    ),
                    (
                        (-0.002740508761501106 - 26.82270480785415j),
                        (1.7035181259573562 - 0.00606963757710623j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (C_0 / 0.1, C_0 / 2.48)


class W(Medium):
    """Tungsten

    Refs:
    * W. S. M. Werner, K. Glantschnig, C. Ambrosch-Draxl.
      Optical constants and inelastic electron-scattering data for 17
      elemental metals, J. Phys Chem Ref. Data 38, 1013-1092 (2009)
    """

    def __init__(self, variant=None):
        super().__init__(name="W")
        self.eps = 1
        self.sigma = 0

        h = HBAR
        self.dispmod = DispersionModel(
            poles=[
                (a / h, c / h)
                for (a, c) in [
                    (
                        (-0.003954896347816251 - 0.1802335610343007j),
                        (1.8921628775430088 + 418.28416384593686j),
                    ),
                    (
                        (-0.012319513432616953 - 0.0052557601183450126j),
                        (1.756799582125415 + 92.88601294383525j),
                    ),
                    (
                        (-5.0741907106627835 - 0.04234993352459908j),
                        (0.3303495961778782 + 36.26799302329413j),
                    ),
                    (
                        (-0.21756967367414215 - 0.9365555173092157j),
                        (0.38496157871304104 + 24.119535838621516j),
                    ),
                    (
                        (-2.6258028910978863 - 2.6236924266577835j),
                        (-0.23193732824781174 + 41.62320103829054j),
                    ),
                ]
            ]
        )
        self.poles = self.dispmod._poles
        self.frequency_range = (C_0 / 0.1, C_0 / 2.48)


class InP(Medium):
    """Indium Phosphide

    Refs:
    * G. D. Pettit and W. J. Turner. Refractive index of InP,
      J. Appl. Phys. 36, 2081 (1965)
    * A. N. Pikhtin and A. D. Yaskov. Disperson of the refractive index of
      semiconductors with diamond and zinc-blende structures,
      Sov. Phys. Semicond. 12, 622-626 (1978)
    * Handbook of Optics, 2nd edition, Vol. 2. McGraw-Hill 1994
    """

    def __init__(self, variant=None):
        super().__init__(name="InP")
        self.eps = 7.255
        self.sigma = 0

        if variant is None:
            variant = "Pettit1965"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Pettit1965" == variant:
            self.dispmod = Sellmeier(
                [
                    (2.316, 0.6263 ** 2),
                    (2.765, 32.935 ** 2),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 10, C_0 / 0.95)


class Ge(Medium):
    """Germanium

    Refs:
    * Icenogle et al.. Refractive indexes and temperature coefficients of
      germanium and silicon Appl. Opt. 15 2348-2351 (1976)
    * N. P. Barnes and M. S. Piltch. Temperature-dependent Sellmeier
      coefficients and nonlinear optics average power limit for germanium
      J. Opt. Soc. Am. 69 178-180 (1979)
    """

    def __init__(self, variant=None):
        super().__init__(name="Ge")
        self.eps = 9.28156
        self.sigma = 0

        if variant is None:
            variant = "Icenogle1976"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Icenogle1976" == variant:
            self.dispmod = Sellmeier(
                [
                    (6.72880, 0.44105),
                    (0.21307, 3870.1),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 12, C_0 / 2.5)


class YAG(Medium):
    """Yttrium aluminium garnet

    Refs:
    * D. E. Zelmon, D. L. Small and R. Page.
      Refractive-index measurements of undoped yttrium aluminum garnet
      from 0.4 to 5.0 um, Appl. Opt. 37, 4933-4935 (1998)
    """

    def __init__(self, variant=None):
        super().__init__(name="YAG")
        self.eps = 1
        self.sigma = 0

        if variant is None:
            variant = "Zelmon1998"

        eV_to_Hz = 0.5 / (math.pi * HBAR)
        if "Zelmon1998" == variant:
            self.dispmod = Sellmeier(
                [
                    (2.28200, 0.01185),
                    (3.27644, 282.734),
                ]
            )
            self.poles = self.dispmod._poles
            self.frequency_range = (C_0 / 5, C_0 / 0.4)



import numpy as np

def dot_product(fields1, fields2, mesh_step):
    """Dot product between the mode and an arbitrary field distribution, 
    in units of power. Or, if the ``Mode`` object is normalized such that 
    (self, self) = 1, then |(self, fields)|^2 is the fraction of 
    the total power carried by the fields that is specifically carried by 
    ``Mode``.
    
    Parameters
    ----------
    fields1 : tuple
        A tuple of (E, H) fields. Each field is array_like of shape 
        ``(3, Nx, Ny)``, with ``z`` assumed to be the propagation direction.
    fields2 : tuple 
        A second tuple of (E, H) fields, with the same shape as ``fields1``.
    mesh_step : tuple
        dx, dy, dz: mesh step in each direction.
    
    Returns
    -------
    float
        The overlap integral between the ``fields1`` and ``fields2``.
    """

    E1 = fields1[0][:2, :, :]
    H1 = fields1[1][:2, :, :]
    E2 = fields2[0][:2, :, :]
    H2 = fields2[1][:2, :, :]

    cross = np.cross(np.conj(E1), H2, axis=0) + \
            np.cross(E2, np.conj(H1), axis=0)
            
    return 1/4*np.sum(cross)*np.prod(mesh_step[:2])


import numpy as np

from .structure import Structure
from .constants import fp_eps, int_, float_, C_0
from .utils import log_and_raise


class Grid(object):
    def __init__(
        self,
        span,
        mesh_step,
        symmetries=[0, 0, 0],
        courant=0.9,
        init_mesh=True,
    ):
        """
        Parameters
        ----------
        span : np.ndarray of shape (3, 2)
            Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the simulation
            region, in micron.
        mesh_step : float or array of floats
            Mesh step in x, y, and z, in micron.
        """

        # Setting span also sets self.mesh, self.mesh_b, self.mesh_f
        self.span = np.array(span).astype(float_)
        self.size = self.span[:, 1] - self.span[:, 0]
        self.set_mesh_step(mesh_step)
        self.symmetries = symmetries

        self.tmesh = None  # To be set when a total simulation time is given
        self.set_time_step(courant)

        def empty_mesh():
            return [np.array([]), np.array([]), np.array([])]

        # Grid size
        self.Nx, self.Ny, self.Nz, self.Nxyz = 0, 0, 0, (0, 0, 0)
        # Backward, centered, and forward meshes
        self.mesh_b, self.mesh, self.mesh_f = [empty_mesh() for i in range(3)]
        # Ex, Ey, and Ez meshes
        self.mesh_ex, self.mesh_ey, self.mesh_ez = [
            empty_mesh() for i in range(3)
        ]
        # H, Hy, and Hz meshes
        self.mesh_hx, self.mesh_hy, self.mesh_hz = [
            empty_mesh() for i in range(3)
        ]

        if init_mesh == True:
            self.init_meshes()

    def set_mesh_step(self, mesh_step):
        ms_tmp = np.array(mesh_step)
        if ms_tmp.size == 1:
            self.mesh_step = ms_tmp * np.ones((3,), dtype=float_)
        elif ms_tmp.size == 3:
            self.mesh_step = ms_tmp
        else:
            log_and_raise(
                "'mesh_step' must be a float or an array of 3 floats.",
                ValueError,
            )

    def init_meshes(self):
        """Initialize centered, forward, and backward meshes. Also initialize 
        the meshes corresponding to the E- and H-field locations on the 
        Yee grid.
        """

        # Slightly increase span to assure pixels at the edges are included.
        _span = self.span
        _span[:, 0] -= fp_eps * self.size
        _span[:, 1] += fp_eps * self.size

        # Initialize mesh points in x, y and z
        self.Nxyz = [
            int_((_span[0][1] - _span[0][0]) / self.mesh_step[0]),
            int_((_span[1][1] - _span[1][0]) / self.mesh_step[1]),
            int_((_span[2][1] - _span[2][0]) / self.mesh_step[2]),
        ]

        # Always take an even number of points if symmetry required
        for dim in range(3):
            if self.symmetries[dim] != 0 and self.Nxyz[dim] % 2 == 1:
                self.Nxyz[dim] += 1
        self.Nx, self.Ny, self.Nz = self.Nxyz

        xcent = (_span[0][1] + _span[0][0]) / 2
        ycent = (_span[1][1] + _span[1][0]) / 2
        zcent = (_span[2][1] + _span[2][0]) / 2

        # Make simulation center coincide with beginning of Yee cell
        xgrid = xcent + self.mesh_step[0] * np.arange(
            -((self.Nx + 1) // 2), (self.Nx) // 2
        )
        ygrid = ycent + self.mesh_step[1] * np.arange(
            -((self.Ny + 1) // 2), (self.Ny) // 2
        )
        zgrid = zcent + self.mesh_step[2] * np.arange(
            -((self.Nz + 1) // 2), (self.Nz) // 2
        )

        for dim, grid in enumerate([xgrid, ygrid, zgrid]):
            # Coordinates of the starting points of the mesh elements
            self.mesh_b[dim] = np.copy(grid).astype(float_)
            # Coordinates of the mesh centers
            self.mesh[dim] = self.mesh_b[dim] + self.mesh_step[dim] / 2
            # Coordinates of the ending points of the mesh elements
            self.mesh_f[dim] = self.mesh_b[dim] + self.mesh_step[dim]

        self.yee_meshes()

    def load_mesh(self, mesh):
        """Load an externally-defined mesh."""
        for dim in range(3):
            self.mesh_b[dim] = mesh[dim] - self.mesh_step[dim] / 2
            self.mesh[dim] = mesh[dim]
            self.mesh_f[dim] = mesh[dim] + self.mesh_step[dim] / 2

        self.Nxyz = [m.size for m in mesh]
        self.Nx, self.Ny, self.Nz = self.Nxyz
        self.yee_meshes()

    def yee_meshes(self):
        # Meshes for the Ex, Ey, and Ez locations.
        self.mesh_ex = (self.mesh[0], self.mesh_b[1], self.mesh_b[2])
        self.mesh_ey = (self.mesh_b[0], self.mesh[1], self.mesh_b[2])
        self.mesh_ez = (self.mesh_b[0], self.mesh_b[1], self.mesh[2])
        self.mesh_hx = (self.mesh_b[0], self.mesh[1], self.mesh[2])
        self.mesh_hy = (self.mesh[0], self.mesh_b[1], self.mesh[2])
        self.mesh_hz = (self.mesh[0], self.mesh[1], self.mesh_b[2])

    def set_time_step(self, stability_factor=0.9):
        """Set the time step based on the generalized Courant stability
        Delta T < 1 / C_0 / sqrt(1 / dx^2 + 1/dy^2 + 1/dz^2)
        dt = courant_condition * stability_factor, so stability factor
        should be < 1.
        """

        dL_sum = np.sum([1 / self.mesh_step[ir] ** 2 for ir in range(3)])
        dL_avg = 1 / np.sqrt(dL_sum)
        courant_stability = dL_avg / C_0
        self.dt = float_(courant_stability * stability_factor)

    def set_tmesh(self, T=0):
        """Set the time mesh for a total simulation time T in seconds."""
        self.tmesh = np.arange(0, T, self.dt)



import sys

import boto3
from .config import Config
keys = Config.user
s3Client = boto3.client(
    's3',
    aws_access_key_id=keys['userAccessKey'],
    aws_secret_access_key=keys['userSecretAccessKey'],
    region_name=Config.S3_REGION
)

class UploadProgress(object):

    def __init__(self, size):
        self.size = size
        self.uploadedSoFar = 0

    def report(self, bytes_in_chunk):
        self.uploadedSoFar += bytes_in_chunk
        sys.stdout.write('\rfile upload progress: {0:2.2f} %'.format(
                        float(self.uploadedSoFar)/self.size*100))
        sys.stdout.flush()


import numpy as np

"""
This file contains constants that are used throghout the codebase.
"""

# Machine precision
dp_eps = np.finfo(np.float64).eps
fp_eps = np.finfo(np.float32).eps

# Very large value
inf = 1e20

# Data types to use
int_ = np.int32
float_ = np.float32
complex_ = np.complex64

# Physical constants. Base unit of length is micron.
EPSILON_0 = np.float32(8.85418782e-18)         # vacuum permittivity [F/um]
MU_0 = np.float32(1.25663706e-12)              # vacuum permeability [H/um]
C_0 = 1 / np.sqrt(EPSILON_0 * MU_0)            # speed of light in vacuum [um/s]
ETA_0 = np.sqrt(MU_0 / EPSILON_0)              # vacuum impedance
Q_e = 1.602176634e-19                          # funamental charge
HBAR = 6.582119569e-16                         # reduced Planck constant [eV*s]

# Conversion between 'x', 'y', 'z' and 0, 1, 2
xyz_dict = {'x': 0, 'y': 1, 'z': 2}
xyz_list = ['x', 'y', 'z']




import numpy as np
import subprocess
import logging

def listify(obj):
    # Make a list if not a list
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return [obj]


def list2str(li, elem_format):
    # Make a string that looks like the list ``li`` using %-specifying string 
    # ``elem_format`` for each element
    return "[" + ", ".join([elem_format % elem for elem in li]) + "]"


def subprocess_cmd(command):
    """Execute a (multi-line) shell command.

    Parameters
    ----------
    command : str
        Semicolon separated lines.
    """
    comm_lines = command.split(";")
    for line in comm_lines:
        comm_list = list(line.split())
        process = subprocess.run(
            comm_list, stdout=None, check=True, stdin=subprocess.DEVNULL
        )


def object_name(obj_list, obj, prefix=""):
    """Return a name for the object that is unique within the names in a
    given list. The optional prefix is to be used if object.name is None.
    """
    if obj.name is not None:
        prefix = obj.name

    name_list = [obj.name for obj in obj_list]
    count = 1
    name = prefix
    if obj.name is None:
        name += "_0"

    while name in name_list:
        name = prefix + "_" + str(count)
        count += 1

    return name


def verbose(print_info):
    if print_info==True:
        logger = logging.getLogger()
        logger.handlers[-1].setLevel(logging.DEBUG)
    elif print_info==False:
        logger = logging.getLogger()
        logger.handlers[-1].setLevel(logging.WARNING)



# class Config(object):
#     S3_REGION="us-east-1"
#     STUDIO_BUCKET="flow360-studio-v1"
#     WEB_API_ENDPONT="https://webapi-dev.flexcompute.com"
#     SOLVER_VERSION="beta-21.1.3.7"

#     # auth info
#     auth = None
#     user = None

#     #other
#     auth_retry = 0

class Config(object):

    S3_REGION="us-gov-west-1"
    STUDIO_BUCKET="flow360studio"
    WEB_API_ENDPONT="https://webapi.flexcompute.com"
    SOLVER_VERSION="release-21.1.3.7"

    # auth info
    auth = None
    user = None

    #other
    auth_retry = 0



import numpy as np
from ..constants import float_
from ..utils import inside_box_inds, listify, log_and_raise
from .solver import compute_modes
from .dot_product import dot_product

class ModePlane(object):
    """2D plane for computation of modes used in ModeSource and ModeMonitor. The 
    coordinate system with respect to which meshes and fields are defined here 
    is always such that the z-axis is normal to the plane.
    """
    def __init__(self, span, norm_ind):
        """Construct.
        
        Parameters
        ----------
        span : np.ndarray of shape (3, 2)
            (micron) Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the 
            mode plane.
        norm_ind : int
            Specifies the normal direction. We must then also have 
            ``span[mode_ind, 0] = span[mode_ind, 1]``.
        """
        self.span = span
        self.norm_ind = norm_ind
        self.cross_inds = [0, 1, 2]
        self.cross_inds.pop(self.norm_ind)

        """Variables below to be set after based on the Simulation object to 
        which the ModePlane is added. """

        # Permittivity and mesh at the center of the Yee cell.
        self.mesh = None
        self.eps = None
        # Permittivity and mesh for field components along the two 
        # cross-section directions and the normal direction
        self.mesh_e1 = None
        self.eps_e1 = None
        self.mesh_e2 = None
        self.eps_e2 = None
        self.mesh_en = None
        self.eps_en = None

        """List of modes, set by a call to ``compute_modes()``. The first list 
        dimension is equal to the number of sampling frequencies, while the 
        second dimension is the number of computed modes. Each mode is given by 
        a dictionary with the fields and propagation constants."""
        self.modes = [[]]
        self.freqs = []

    def _set_eps(self, sim):
        """ Set the mesh of the ModePlane based on a global simulation mesh, 
        and the indexing of that mesh w.r.t. the simulation grid.
        """
        indsx, indsy, indsz = inside_box_inds(self.span, sim.grid.mesh,
                                            include_zero_size=True)
        if np.any([inds[0]==inds[1] for inds in (indsx, indsy, indsz)]):
            log_and_raise(
                "Mode plane position is outside simulation domain.",
                ValueError
            )

        """Array of shape (3, 2) of int defining the starting and stopping 
        index in the global simulation grid of the ModePlane span."""
        self.span_inds = np.array([[inds[0], inds[1]] 
                                    for inds in (indsx, indsy, indsz)])

        # Space and time resolution from global grid.
        self.mesh_step = (sim.grid.mesh_step[self.cross_inds[0]], 
                     sim.grid.mesh_step[self.cross_inds[1]],
                     sim.grid.mesh_step[self.norm_ind])
        self.time_step = sim.grid.dt

        meshes = []
        epses = []
        sim_ms = [sim.grid.mesh, sim.grid.mesh_ex,
                    sim.grid.mesh_ey, sim.grid.mesh_ez]
        for im, mesh in enumerate(sim_ms):
            # Average the edges for the Yee grid coordinates, otherwise 
            # include them in eps (used for plotting)
            edges = 'average' if im > 0 else 'in'
            plane_mesh = (mesh[0][indsx[0]:indsx[1]],
                          mesh[1][indsy[0]:indsy[1]],
                          mesh[2][indsz[0]:indsz[1]])
            meshes.append([plane_mesh[self.cross_inds[0]], 
                          plane_mesh[self.cross_inds[1]],
                          plane_mesh[self.norm_ind]])
            eps = sim._get_eps(plane_mesh, edges=edges)
            if np.prod(eps.shape)== 0 or np.count_nonzero(eps.shape==1) > 1:
                log_and_raise(
                    "Mode plane cross-section size too small.",
                    ValueError
                )
            epses.append(np.squeeze(eps))

        self.mesh = meshes[0]
        self.mesh_e1 = meshes[1 + self.cross_inds[0]]
        self.mesh_e2 = meshes[1 + self.cross_inds[1]]
        self.mesh_en = meshes[1 + self.norm_ind]
        self.eps = epses[0]
        self.eps_e1 = epses[1 + self.cross_inds[0]]
        self.eps_e2 = epses[1 + self.cross_inds[1]]
        self.eps_en = epses[1 + self.norm_ind]

    def compute_modes(self, Nmodes, freqs, pml_layers=[0, 0]):
        """ Compute the ``Nmodes`` eigenmodes in decreasing order of 
        propagation constant at every frequency in the list ``freqs``.
        """

        if self.eps is None:
            raise ValueError("ModePlane object has not been added to a "
                            "simulation yet.")

        # Add a little bit of anisotropy to break potential degeneracy
        epses = [self.eps_e1, self.eps_e2 + 1e-6, self.eps_en]
        self.freqs = listify(freqs)
        self.modes = []
        for frequency in self.freqs:
            # Get modes
            modes = compute_modes(epses, frequency,
                                    mesh_step=self.mesh_step,
                                    pml_layers=pml_layers,
                                    num_modes=Nmodes)
            
            for mode in modes:
                # Normalize to unit power flux
                fields_cent = mode.fields_to_center()
                flux = dot_product(fields_cent, fields_cent, self.mesh_step)
                mode.E /= np.sqrt(flux)
                mode.H /= np.sqrt(flux)
                # Make largest E-component real
                mode.fix_efield_phase()

            self.modes.append(modes)


import numpy as np
from matplotlib.path import Path

# gdspy is only needed for the GdsSlab structure
try:
    import gdspy
except ImportError:
    pass

from .material import Medium, PEC, PMC
from .utils import inside_box, cs2span, log_and_raise
from .constants import fp_eps, int_, float_, xyz_dict

class Structure(object):

    def __init__(self, material, name=None):
        """Base class for structures. Available subclasses:

        - :class:`.Box`
        - :class:`.Sphere`
        - :class:`.Cylinder`
        - :class:`.PolySlab`
        - :class:`.GdsSlab`
        """
        self.material = material
        self.name = str(name) if name else None

        # if bounds are not set (default), self.bounds is None,
        # otherwise it's a (3,2) array specifying bounding box min and max
        # values along each dimension
        self.bounds = None

    def _inside(self, mesh, include_edges=True):
        """Elementwise indicator function for the structure.
        
        Parameters
        ----------
        mesh : tuple
            3-tuple defining the xgrid, ygrid and zgrid.
        include_edges : bool
            Whether a point sitting exactly on a mesh point (within numerical 
            precision) should be returned as inside (True) or outside (False) 
            the structure.
        
        Note
        ----
        ``include_edges`` will in the future be replaced by actual dielectric 
        smoothening.
        
        Returns
        -------
        mask : np.ndarray
            A 3D array of shape (mesh[0].size, mesh[1].size, mesh[2].size) 
            that is 1 inside the structure and 0 outside, and a continuous 
            value between 0 and 1 at interfaces if smoothen==True.
        """

        raise NotImplementedError(
            "inside() needs to be implemented by Structure subclasses"
        )

    def _get_eps_val(self, pec_val, pmc_val):
        """ get epsilon value for structure """

        if isinstance(self.material, Medium):
            return self.material.eps
        elif isinstance(self.material, PEC):
            return pec_val
        elif isinstance(self.material, PMC):
            return pmc_val

    def _set_val_direct(self, mesh, val_arr, val, edges='average'):
        """ Set value ``val`` to all elements of array ``val_arr`` defined 
        over sptial ``mesh``, which are inside the current Structure."""

        if edges=='in' or edges=='average':
            mask = self._inside(mesh, include_edges=True)
        if edges=='average':
            mask += self._inside(mesh, include_edges=False)
            mask = mask.astype(float_)/2
        if edges=='out':
            mask = self._inside(mesh, include_edges=False)
        mask_bool = mask > 0
        mask_in = mask[mask_bool]
        val_arr[mask_bool] = (1 - mask_in)*val_arr[mask_bool] + val*mask_in

    def _set_val(self, mesh, val_arr, val, edges='average'):
        """ Set value ``val`` to all elements of array ``val_arr`` defined 
        over sptial ``mesh``, which are inside the current Structure. This 
        uses ``_set_val_direct`` only after applying the bounding box, if 
        known."""

        # no bounds defined, do the most naive thing
        if self.bounds is None:
            self._set_val_direct(mesh, val_arr, val, edges=edges)
            return

        bound_inds = self._get_bounding_indices(mesh)

        # mesh is completely outside bounds, skip this structure
        if bound_inds is None:
            return

        sub_mesh, sub_arr = self._get_sub_problem(mesh, val_arr, bound_inds)
        self._set_val_direct(sub_mesh, sub_arr, val, edges=edges)

    def _get_bounding_indices(self, mesh):
        """ uses self.bounds to compute min and max bounding indices in mesh
        
        Parameters
        ----------
        mesh : tuple
            3-tuple defining the xgrid, ygrid and zgrid.
        
        Returns
        -------
        bound_inds : np.ndarray
            A 3D array of shape (3, 2) 
            bound_inds[d, :] gives the indices into `mesh` that are closest to 
            min and max values in dimension `d`. 
            for safety, min is rounded down to the nearest index and 
            max is rounded up to the nearest index
        """

        # if bounds are not defined for this structure, just return None
        if self.bounds is None:
            return None

        # unpack bounds and mesh
        xmin, xmax = self.bounds[0]
        ymin, ymax = self.bounds[1]
        zmin, zmax = self.bounds[2]
        xs, ys, zs = mesh

        # compute indices within the bounding region in each dimension
        ixs_min = np.where(xs > xmin)[0]
        ixs_max = np.where(xs < xmax)[0]
        iys_min = np.where(ys > ymin)[0]
        iys_max = np.where(ys < ymax)[0]
        izs_min = np.where(zs > zmin)[0]
        izs_max = np.where(zs < zmax)[0]

        # check if any conditions gave no matches (bounds outside mesh, skip)
        indices = (ixs_min, ixs_max, iys_min, iys_max, izs_min, izs_max)
        if any([len(i) == 0 for i in indices]):
            return None

        # get the min and max inside bounding indices in each dimension
        ix_min, ix_max = np.min(ixs_min), np.max(ixs_max) + 1
        iy_min, iy_max = np.min(iys_min), np.max(iys_max) + 1
        iz_min, iz_max = np.min(izs_min), np.max(izs_max) + 1

        # construct bounding indices array for constructing sub mesh & sub arr
        return np.array([[ix_min, ix_max], [iy_min, iy_max], [iz_min, iz_max]])

    @staticmethod
    def _get_sub_problem(mesh, arr, bound_inds):
        """ returns a view to sub mesh and sub epsilon array using `bound_inds`
        """

        # unpack bound_inds
        bounds_ix, bounds_iy, bounds_iz = bound_inds
        ix_min, ix_max = bounds_ix
        iy_min, iy_max = bounds_iy
        iz_min, iz_max = bounds_iz

        # get sub mesh
        xs, ys, zs = mesh
        sub_xs = xs[ix_min:ix_max]
        sub_ys = ys[iy_min:iy_max]
        sub_zs = zs[iz_min:iz_max]
        sub_mesh = sub_xs, sub_ys, sub_zs

        # get view into sub epsilon array
        sub_eps = arr[ix_min:ix_max, iy_min:iy_max, iz_min:iz_max]

        return sub_mesh, sub_eps


class Box(Structure):
    """ Box structure, i.e. a 3D rectangular block.
    """

    def __init__(self, center, size, material, name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron): x, y, and z position of the center of the Box.
        size : array_like
            (micron): size in x, y, and z.
        material : Material
            Material of the structure.
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.center = np.array(center)
        self.size = np.array(size)
        self.span = cs2span(self.center, self.size)

        # set bounds
        mins = self.center - self.size / 2 / (1. - fp_eps)
        maxs = self.center + self.size / 2 / (1. - fp_eps)
        self.bounds = np.stack((mins, maxs), axis=1)  # (3, 2)

    def _inside(self, mesh, include_edges=True):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the Box region."""

        # note: there's a bug here if self.size is
        tmp_span = np.copy(self.span)
        if include_edges == True:
            tmp_span[:, 0] -= fp_eps*(tmp_span[:, 1] - tmp_span[:, 0])
            tmp_span[:, 1] += fp_eps*(tmp_span[:, 1] - tmp_span[:, 0])
        else:
            tmp_span[:, 0] += fp_eps*(tmp_span[:, 1] - tmp_span[:, 0])
            tmp_span[:, 1] -= fp_eps*(tmp_span[:, 1] - tmp_span[:, 0])

        return inside_box(tmp_span, mesh, include_zero_size=False)


class BoxAngled(Structure):
    """ Box structure, i.e. a 3D rectangular block at an angle in x-y plane
    """

    def __init__(self, center, size, angle, material, bounded=True, name=None):
        """ Box positioned at angle in x-y plane
        Parameters
        ----------
        center : array_like
            (micron): x, y, and z position of the center of the Box.
        size : array_like
            (micron): size in x, y, and z.
        angle: float
            (radians) angle to rotate box from x axis
        material : Material
            Material of the structure.
        bounded : bool
            If you want to use the bounded feature 
            (just for testing with and without)
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.center = np.array(center)
        self.size = np.array(size)
        self.angle = angle
        if bounded:
            xy_radius = np.max(self.size[:-1])
            z_size = self.size[-1]
            max_xy_size = 2 * xy_radius * np.sqrt(2)
            size = np.array([max_xy_size, max_xy_size, z_size])
            mins = self.center - size / 2 / (1. - fp_eps)
            maxs = self.center + size / 2 / (1. - fp_eps)
            self.bounds = np.stack((mins, maxs), axis=1)  # (3, 2)

    def _inside(self, mesh, include_edges=False):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the Box region."""

        # expand mesh
        xx, yy, zz = np.meshgrid(*mesh, indexing="ij")

        # centered coordinates
        xxc = xx - self.center[0]
        yyc = yy - self.center[1]
        zzc = zz - self.center[2]

        # rotated coords
        xxp = xxc * np.cos(self.angle) - yyc * np.sin(self.angle)
        yyp = xxc * np.sin(self.angle) + yyc * np.cos(self.angle)

        # return 1 if rotated coordinates are inside of box
        inside_x = np.abs(xxp) < self.size[0] / 2
        inside_y = np.abs(yyp) < self.size[1] / 2
        inside_z = np.abs(zzc) < self.size[2] / 2

        return 1.0 * (inside_x * inside_y * inside_z)


class Sphere(Structure):
    """ Sphere structre.
    """

    def __init__(self, center, radius, material, name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron): x, y, z position of the center of the sphere.
        radius : float
            (micron) Radius of the sphere.
        material : Material
            Material of the structure.
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.center = np.array(center, dtype=float_)
        self.radius = radius
        mins = self.center - 2 * self.radius / (2. - fp_eps)
        maxs = self.center + 2 * self.radius / (2. - fp_eps)
        self.bounds = np.stack((mins, maxs), axis=1)

    def _inside(self, mesh, include_edges=True):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the Sphere."""

        xc, yc, zc = self.center

        # this line is abstruse
        r = self.radius * (1 + (include_edges - 0.5) * 2 * fp_eps)

        # cleaned this up:
        #
        # return np.where((x - mesh[0][:, np.newaxis, np.newaxis])**2 +
        #             (y - mesh[1][np.newaxis, :, np.newaxis])**2 +
        #             (z - mesh[2][np.newaxis, np.newaxis, :])**2 < r**2, 1, 0)
        # becomes:

        xx, yy, zz = np.meshgrid(*mesh, indexing="ij")
        return 1.0 * ((xx - xc) ** 2 + (yy - yc) ** 2 + (zz - zc) ** 2 < r ** 2)


class Cylinder(Structure):
    """ Cylinder structre.
    """

    def __init__(self, center, axis, radius, height, material, name=None):
        """Construct.

        Parameters
        ----------
        center : array_like
            (micron): x, y, z position of the center of the sphere.
        axis : str
            ``'x'``, ``'y'``, or ``'z'``.
        radius : float
            (micron) Radius of the cylinder.
        height : float
            (micron) Height of the cylinder along its axis.
        material : Material
            Material of the structure.
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.center = np.array(center, dtype=float_)
        self.axis = axis
        self.radius = radius
        self.height = height

        # set bounds
        if axis == "x":
            sizes = np.array([height, 2 * radius, 2 * radius])
        elif axis == "y":
            sizes = np.array([2 * radius, height, 2 * radius])
        elif axis == "z":
            sizes = np.array([2 * radius, 2 * radius, height])
        else:
            # do this error checking elsewhere
            log_and_raise(
                f"Given axis {axis}, must be in ['x', 'y', 'z'].",
                ValueError
            )

        mins = self.center - sizes / 2 / (1. - fp_eps)
        maxs = self.center + sizes / 2 / (1. - fp_eps)
        self.bounds = np.stack((mins, maxs), axis=1)  # (3, 2)

    def _inside(self, mesh, include_edges=True):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the Cylinder."""

        ax = xyz_dict[self.axis]
        d_cross = [0, 1, 2]
        d_cross.pop(ax)
        d_a = self.center[ax]
        d1, d2 = self.center[d_cross]
        r = self.radius * (1 + (include_edges - 0.5) * 2 * fp_eps)
        h = self.height * (1 + (include_edges - 0.5) * 2 * fp_eps)

        m = [
            mesh[0][:, np.newaxis, np.newaxis],
            mesh[1][np.newaxis, :, np.newaxis],
            mesh[2][np.newaxis, np.newaxis, :],
        ]
        m_a = m[ax]
        m1 = m[d_cross[0]]
        m2 = m[d_cross[1]]

        return np.where(((d1 - m1)**2 + (d2 - m2)**2 < r**2) *  
                            (np.abs(d_a - m_a) < h/2), 1, 0)


class PolySlab(Structure):
    """ A structure defined as polygon in x and y with a constnant extent in z.
    """

    def __init__(self, vertices, z_cent, z_size, material, name=None):
        """ Construct.

        Parameters
        ----------
        vertices : array_like
            (micron) Shape (N, 2) defining the polygon vertices in the xy-plane.
        z_cent : float
            (micron) Center of the polygonal slab in z.
        z_size : float
            (micron) Thickness of the slab in z.
        material : Material
            Material of the structure.
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.vertices = np.array(vertices, dtype=float_)
        self.z_cent = z_cent
        self.z_size = z_size

        x_min, y_min = np.min(vertices, axis=0)
        x_max, y_max = np.max(vertices, axis=0)
        z_min = z_cent - z_size / 2
        z_max = z_cent + z_size / 2

        mins = np.array([x_min, y_min, z_min])
        maxs = np.array([x_max, y_max, z_max])
        self.bounds = np.stack((mins, maxs), axis=1)  # (3, 2)

    def _inside(self, mesh, include_edges=True):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the PolySlab."""

        z_size = self.z_size * (1 + (include_edges - 0.5) * 2 * fp_eps)

        path = Path(self.vertices)
        xm, ym = np.meshgrid(mesh[0], mesh[1])
        points = np.vstack((xm.ravel(), ym.ravel())).T
        mask2d = path.contains_points(points).reshape(xm.shape).T
        maskz = (mesh[2] >= self.z_cent - z_size / 2) * (
            mesh[2] <= self.z_cent + z_size / 2
        )

        return np.where(
            mask2d[:, :, np.newaxis] * maskz[np.newaxis, np.newaxis, :] > 0, 1, 0
        )


class GdsSlab(Structure):
    """ A structure defined through a GDSII cell imported through ``gdspy``. 
    All polygons and paths included in the cell are assumed to lie in the 
    xy-plane, with the same center and size in z, and made of the same material.
    """

    def __init__(self, gds_cell, z_cent, z_size, material, name=None):
        """Construct.

        Parameters
        ----------
        gds_cell : gdspy.Cell
            A GDS Cell containing all 2D polygons and paths, in microns.
        z_cent : float
            (micron) Center of the slab(s) in z.
        z_size : float
            (micron) Thickness of the slab(s) in z.
        material : Material
            Material of the structure.
        name : str, optional
            Custom name of the structure.
        """
        super().__init__(material, name)
        self.gds_cell = gds_cell
        self.z_cent = z_cent
        self.z_size = z_size

        vertices = gds_cell.get_polygons()
        xy_min = np.min(vertices, axis=0)
        xy_max = np.max(vertices, axis=0)
        z_min = z_cent - z_size / 2 / (1. - fp_eps)
        z_max = z_cent + z_size / 2 / (1. - fp_eps)

        # GDSpy already does bounds (?)
        # mins = np.array([xy_min, xy_min, z_min])
        # maxs = np.array([xy_max, xy_max, z_max])
        # self.bounds = np.stack((mins, maxs), axis=1) # (3, 2)

    def _inside(self, mesh, include_edges=True):
        """Returns a mask defining whether the points in ``mesh`` are inside 
        the PolySlab."""

        z_size = self.z_size * (1 + (include_edges - 0.5) * 2 * fp_eps)

        xm, ym = np.meshgrid(mesh[0], mesh[1])
        points = np.vstack((xm.ravel(), ym.ravel())).T
        mask2d = (
            np.array(gdspy.inside(points, self.gds_cell.get_polygons()))
            .reshape(xm.shape)
            .T
        )
        maskz = (mesh[2] >= self.z_cent - z_size / 2) * (
            mesh[2] <= self.z_cent + z_size / 2
        )

        return np.where(
            mask2d[:, :, np.newaxis] * maskz[np.newaxis, np.newaxis, :] > 0, 1, 0
        )
